import imspy_connector
ims = imspy_connector.py_chemistry


def calculate_mz(mass: float, charge: int) -> float:
    """Calculate m/z value.

    Args:
        mass (float): Mass.
        charge (int): Charge.

    Returns:
        float: m/z value.
    """
    return ims.calculate_mz(mass, charge)




import re
from typing import List


def tokenize_unimod_sequence(unimod_sequence: str) -> List[str]:
    """
    Tokenizes a sequence of modified amino acids.
    Args:
        unimod_sequence: A string representing the sequence of amino acids with modifications.

    Returns:
        A list of tokenized amino acids.
    """
    token_pattern = r'[A-Z](?:\[UNIMOD:\d+\])?'

    # Special case handling for [UNIMOD:1] at the beginning
    if unimod_sequence.startswith("[UNIMOD:1]"):
        special_token = "<START>[UNIMOD:1]"
        rest_of_string = unimod_sequence[len("[UNIMOD:1]"):]
        other_tokens = re.findall(token_pattern, rest_of_string)
        return [special_token] + other_tokens + ['<END>']
    else:
        tokens = re.findall(token_pattern, unimod_sequence)
        return ['<START>'] + tokens + ['<END>']



import sqlite3
import pandas as pd

from imspy.timstof.data import TimsDataset
from imspy.timstof.frame import TimsFrame

import imspy_connector
ims = imspy_connector.py_dda


class TimsDatasetDDA(TimsDataset):

    def __init__(self, data_path: str):
        super().__init__(data_path=data_path)
        self.__dataset = ims.PyTimsDatasetDDA(self.data_path, self.binary_path)
        self.meta_data = self.meta_data.rename(columns={"Id": "frame_id"})
        self.fragmented_precursors = self._load_selected_precursors().rename(
            columns={
                'Id': 'precursor_id',
                'LargestPeakMz': 'largest_peak_mz',
                'AverageMz': 'average_mz',
                'MonoisotopicMz': 'monoisotopic_mz',
                'Charge': 'charge',
                'ScanNumber': 'average_scan',
                'Intensity': 'intensity',
                'Parent': 'parent_id',
            }
        )
        self.pasef_meta_data = self._load_pasef_meta_data().rename(
            columns={
                'Frame': 'frame_id',
                'ScanNumBegin': 'scan_begin',
                'ScanNumEnd': 'scan_end',
                'IsolationMz': 'isolation_mz',
                'IsolationWidth': 'isolation_width',
                'CollisionEnergy': 'collision_energy',
                'Precursor': 'precursor_id'
            }
        )

    def _load_selected_precursors(self):
        """Get precursors selected for fragmentation.

        Returns:
            pd.DataFrame: Precursors selected for fragmentation.
        """
        return pd.read_sql_query("SELECT * from Precursors", sqlite3.connect(self.data_path + "/analysis.tdf"))

    def _load_pasef_meta_data(self):
        """Get PASEF meta data for DDA.

        Returns:
            pd.DataFrame: PASEF meta data.
        """
        return pd.read_sql_query("SELECT * from PasefFrameMsMsInfo",
                                 sqlite3.connect(self.data_path + "/analysis.tdf"))

    def get_pasef_fragments(self) -> pd.DataFrame:
        """Get PASEF fragments.

        Args:
            num_threads (int, optional): Number of threads. Defaults to 4.

        Returns:
            List[FragmentDDA]: List of PASEF fragments.
        """
        pasef_fragments = [FragmentDDA.from_py_tims_fragment_dda(fragment)
                           for fragment in self.__dataset.get_pasef_fragments(1)]

        pasef_fragments = pd.DataFrame({
            'frame_id': [s.frame_id for s in pasef_fragments],
            'precursor_id': [s.precursor_id for s in pasef_fragments],
            'raw_data': [s.selected_fragment for s in pasef_fragments]
        })

        A = pd.merge(
            pasef_fragments, self.pasef_meta_data,
            left_on=['precursor_id', 'frame_id'],
            right_on=['precursor_id', 'frame_id'],
            how='inner',
        )

        B = pd.merge(
            A, self.fragmented_precursors,
            left_on=['precursor_id'],
            right_on=['precursor_id'],
            how='inner'
        )

        time = self.meta_data[['frame_id']]
        time.insert(time.shape[1], "time", self.meta_data['Time'] / 60)
        
        return pd.merge(time, B, left_on=['frame_id'], right_on=['frame_id'], how='inner')

    def __repr__(self):
        return (f"TimsDatasetDDA(data_path={self.data_path}, num_frames={self.frame_count}, "
                f"fragmented_precursors={self.fragmented_precursors.shape[0]})")


class FragmentDDA:
    def __init__(self, frame_id: int, precursor_id: int, selected_fragment: TimsFrame):
        self._fragment_ptr = ims.PyTimsFragmentDDA(frame_id, precursor_id, selected_fragment.get_frame_ptr())

    @classmethod
    def from_py_tims_fragment_dda(cls, fragment: ims.PyTimsFragmentDDA):
        instance = cls.__new__(cls)
        instance._fragment_ptr = fragment
        return instance

    @property
    def frame_id(self) -> int:
        return self._fragment_ptr.frame_id

    @property
    def precursor_id(self) -> int:
        return self._fragment_ptr.precursor_id

    @property
    def selected_fragment(self) -> TimsFrame:
        return TimsFrame.from_py_tims_frame(self._fragment_ptr.selected_fragment)

    def __repr__(self):
        return f"FragmentDDA(frame_id={self.frame_id}, precursor_id={self.precursor_id}, " \
               f"selected_fragment={self.selected_fragment})"

    def get_fragment_ptr(self):
        return self._fragment_ptr



""" This module contains several noise models,
    such as detection noise, shot noise and baseline noise.
"""
import numpy as np
import numba
from typing import Callable, Optional
from numpy.typing import ArrayLike
from imspy.data.spectrum import MzSpectrum
from imspy.utility.utilities import normal_pdf


@numba.jit(nopython=True)
def mu_function_normal_default(intensity: ArrayLike) -> ArrayLike:
    return intensity


@numba.jit(nopython=True)
def sigma_function_normal_default(intensity:ArrayLike) -> ArrayLike:
    return np.sqrt(intensity)


@numba.jit(nopython=True)
def mu_function_poisson_default(intensity:ArrayLike) -> ArrayLike:
    offset = 0
    return intensity + offset


@numba.jit(nopython=True)
def detection_noise(signal: ArrayLike,
                    method: str = "poisson",
                    custom_mu_function: Optional[Callable] = None,
                    custom_sigma_function: Optional[Callable] = None) -> ArrayLike:
    """

    :param signal:
    :param method:
    :param custom_mu_function:
    :param custom_sigma_function:
    """
    if method == "normal":
        if custom_sigma_function is None:
            sigma_function:Callable = sigma_function_normal_default
        else:
            sigma_function:Callable = custom_sigma_function
        if custom_mu_function is None:
            mu_function:Callable = mu_function_normal_default
        else:
            mu_function:Callable = custom_mu_function

        sigmas = sigma_function(signal)
        mus = mu_function(signal)
        noised_signal = np.zeros_like(mus)

        for i in range(noised_signal.size):
            s = sigmas[i]
            m = mus[i]
            n = np.random.normal()*s+m
            noised_signal[i] = n

    elif method == "poisson":
        if custom_sigma_function is not None:
            raise ValueError("Sigma function is not used if method is 'poisson'")
        if custom_mu_function is None:
            mu_function:Callable = mu_function_poisson_default
        else:
            mu_function:Callable = custom_mu_function
        mus = mu_function(signal)
        noised_signal = np.zeros_like(mus)

        for i in range(noised_signal.size):
            mu = mus[i]
            n = np.random.poisson(lam = mu)
            noised_signal[i] = n

    else:
        raise NotImplementedError("This method is not implemented, choose 'normal' or 'poisson'.")

    return noised_signal


@numba.jit(nopython=True)
def generate_noise_peak(pos:float, sigma: float, intensity: float, min_intensity:int = 0, resolution:float = 3):
    step_size = min(sigma/10,1/np.power(10,resolution))
    lower = int((pos-4*sigma)//step_size)
    upper = int((pos+4*sigma)//step_size)
    mzs = np.arange(lower,upper)*step_size
    intensities = normal_pdf(mzs,pos,sigma,normalize=True)*intensity*step_size
    return (mzs[intensities>=min_intensity], intensities[intensities>=min_intensity].astype(np.int64))


def baseline_shot_noise_window(window:MzSpectrum,
                               window_theoretical_mz_min:float,
                               window_theoretical_mz_max:float,
                               expected_noise_peaks: int = 5,
                               expected_noise_intensity: float = 10,
                               expected_noise_sigma:float = 0.001,
                               resolution:float = 3) -> MzSpectrum:
    """

    """
    num_noise_peaks = np.random.poisson(lam=expected_noise_peaks)
    noised_window = MzSpectrum([],[])
    for i in range(num_noise_peaks):
        location_i = np.random.uniform(window_theoretical_mz_min,window_theoretical_mz_max)
        intensity_i = np.random.exponential(expected_noise_intensity)
        sigma_i = np.random.exponential(expected_noise_sigma)
        noise_mz, noise_intensity = generate_noise_peak(location_i,sigma_i,intensity_i,resolution=resolution)
        noise_peak = MzSpectrum(noise_mz, noise_intensity)
        noised_window = noised_window+noise_peak
    return noised_window


def baseline_shot_noise(spectrum:MzSpectrum,window_size:float=50,expected_noise_peaks_per_Th:int=10,min_intensity:int = 5, resolution = 3):
    """


    """
    min_mz = spectrum.mz.min()-0.1
    max_mz = spectrum.mz.max()+0.1
    bins,windows = spectrum.windows(window_size,overlapping=False,min_peaks=0,min_intensity=0)
    noised_spectrum = spectrum
    for b,w in zip(bins,windows):
        lower = b*window_size
        upper = (b+1)*window_size
        noised_spectrum = noised_spectrum+baseline_shot_noise_window(w,lower,upper,window_size*expected_noise_peaks_per_Th)
    return noised_spectrum.to_resolution(resolution).filter(min_mz,max_mz,min_intensity)


def baseline_noise():
    pass



from __future__ import annotations
from abc import ABC, abstractmethod
from typing import List, Tuple
from multiprocessing import Pool

import tensorflow as tf
import numpy as np
from numpy.typing import ArrayLike, NDArray
import pandas as pd
from scipy.stats import exponnorm, norm, binom, gamma

from imspy.chemistry.mass import STANDARD_TEMPERATURE, STANDARD_PRESSURE, BufferGas, get_num_protonizable_sites
from imspy.chemistry.mobility import SUMMARY_CONSTANT as CCS_K0_CONVERSION_CONSTANT
from imspy.simulation.proteome import ProteomicsExperimentSampleSlice
from imspy.simulation.feature import RTProfile, ScanProfile, ChargeProfile
from imspy.simulation.isotopes import AveragineGenerator
from imspy.utility.utilities import tokenizer_from_json


class Device(ABC):
    def __init__(self, name: str):
        self.name = name
        self._temperature = STANDARD_TEMPERATURE
        self._pressure = STANDARD_PRESSURE

    @property
    def temperature(self):
        """
        Get device temperature

        :return: Temperature of device in Kelvin.
        :rtype: float
        """
        return self._temperature

    @temperature.setter
    def temperature(self, T:float):
        """
        Set device temperature

        :param T: Temperature in Kelvin.
        :type T: float
        """
        self._temperature = T

    @property
    def pressure(self):
        """
        Get device pressure

        :return: Pressure of device in Pa.
        :rtype: float
        """
        return self._pressure

    @pressure.setter
    def pressure(self, p:float):
        """
        Set device pressure
        :param p: Pressure in Pa.
        :type p: float
        """
        self._pressure = p

    @abstractmethod
    def run(self, sample: ProteomicsExperimentSampleSlice):
        pass


class Model(ABC):
    def __init__(self):
        pass

    @abstractmethod
    def simulate(self, sample: ProteomicsExperimentSampleSlice, device: Device):
        pass


class Chromatography(Device):
    def __init__(self, name: str="ChromatographyDevice"):
        super().__init__(name)
        self._apex_model = None
        self._profile_model = None
        self._irt_to_rt_converter = None
        self._frame_length = 1200
        self._gradient_length = 120 * 60 * 1000 # 120 minutes in miliseconds

    @property
    def frame_length(self):
        return self._frame_length

    @frame_length.setter
    def frame_length(self, milliseconds: int):
        self._frame_length = milliseconds

    @property
    def gradient_length(self):
        return self._gradient_length/(60*1000)

    @gradient_length.setter
    def gradient_length(self, minutes: int):
        self._gradient_length = minutes*60*1000

    @property
    def num_frames(self):
        return self._gradient_length//self._frame_length

    @property
    def apex_model(self):
        return self._apex_model

    @apex_model.setter
    def apex_model(self, model: ChromatographyApexModel):
        self._apex_model = model

    @property
    def profile_model(self):
        return self._profile_model

    @property
    def irt_to_rt_converter(self):
        return self._irt_to_rt_converter

    @irt_to_rt_converter.setter
    def irt_to_rt_converter(self, converter:callable):
        self._irt_to_rt_converter = converter

    @profile_model.setter
    def profile_model(self, model:ChromatographyProfileModel):
        self._profile_model = model

    def irt_to_frame_id(self, irt: float):
        return self.rt_to_frame_id(self.irt_to_rt(irt))

    @abstractmethod
    def rt_to_frame_id(self, rt: float):
        pass

    def irt_to_rt(self, irt):
        return self.irt_to_rt_converter(irt)

    def frame_time_interval(self, frame_id:ArrayLike):
        frame_id = np.atleast_1d(frame_id)
        frame_length_minutes = self.frame_length/(60*1000)
        s = (frame_id-1)*frame_length_minutes
        e = frame_id*frame_length_minutes
        return np.stack([s, e], axis = 1)

    def frame_time_middle(self, frame_id: ArrayLike):
        return np.mean(self.frame_time_interval(frame_id),axis=1)

    def frame_time_end(self, frame_id: ArrayLike):
        return self.frame_time_interval(frame_id)[:,1]

    def frame_time_start(self, frame_id: ArrayLike):
        return self.frame_time_interval(frame_id)[:,0]


class LiquidChromatography(Chromatography):
    def __init__(self, name: str = "LiquidChromatographyDevice"):
        super().__init__(name)

    def run(self, sample: ProteomicsExperimentSampleSlice):
        # retention time apex simulation
        retention_time_apex = self._apex_model.simulate(sample, self)
        # in irt and rt
        sample.add_simulation("simulated_irt_apex", retention_time_apex)
        # in frame id
        sample.add_simulation("simulated_frame_apex", self.irt_to_frame_id(retention_time_apex))

        # profile simulation

        retention_profile = self._profile_model.simulate(sample, self)
        sample.add_simulation("simulated_frame_profile", retention_profile)

    def rt_to_frame_id(self,  rt_minutes: ArrayLike):
        rt_minutes = np.asarray(rt_minutes)
        # first frame is completed not at 0 but at frame_length
        frame_id = (rt_minutes/self.frame_length*1000*60).astype(np.int64)+1
        return frame_id


class ChromatographyApexModel(Model):
    def __init__(self):
        self._device = None

    @abstractmethod
    def simulate(self, sample: ProteomicsExperimentSampleSlice, device: Chromatography) -> NDArray[np.float64]:
        pass


class ChromatographyProfileModel(Model):
    def __init__(self):
        pass

    @abstractmethod
    def simulate(self, sample: ProteomicsExperimentSampleSlice, device: Chromatography) -> List[RTProfile]:
        pass


class EMGChromatographyProfileModel(ChromatographyProfileModel):

    def __init__(self):
        super().__init__()

    def simulate(self, sample: ProteomicsExperimentSampleSlice, device: Chromatography) -> List[RTProfile]:
        mus = sample.peptides["simulated_irt_apex"].values
        profile_list = []

        for mu in mus:
            σ = 0.1 # must be sampled
            λ = 10 # must be sampled
            K = 1/(σ*λ)
            μ = device.irt_to_rt(mu)
            model_params = { "sigma":σ,
                             "lambda":λ,
                             "mu":μ,
                             "name":"EMG"
                            }

            emg = exponnorm(loc=μ, scale=σ, K = K)
            # start and end value (in retention times)
            s_rt, e_rt = emg.ppf([0.01,0.9])
            # as frames
            s_frame, e_frame = device.rt_to_frame_id(s_rt), device.rt_to_frame_id(e_rt)

            profile_frames = np.arange(s_frame-1,e_frame+1) # starting with s_frame-1 for cdf interval calculation
            profile_rt_ends = device.frame_time_end(profile_frames)
            profile_rt_cdfs = emg.cdf(profile_rt_ends)
            profile_rel_intensities = np.diff(profile_rt_cdfs)

            profile_list.append(RTProfile(profile_frames[1:],profile_rel_intensities,model_params))
        return profile_list


class NormalChromatographyProfileModel(ChromatographyProfileModel):

    def __init__(self):
        super().__init__()

    def simulate(self, sample: ProteomicsExperimentSampleSlice, device: Chromatography) -> List[RTProfile]:
        mus = sample.peptides["simulated_irt_apex"].values
        sigmas = gamma(a=4.929,scale=1/18.784).rvs(mus.size)/6
        profile_list = []

        for mu,σ in zip(mus,sigmas):

            μ = device.irt_to_rt(mu)
            model_params = { "sigma":σ,
                             "mu":μ,
                             "name":"Normal"
                            }

            normal = norm(loc=μ, scale=σ)
            # start and end value (in retention times)
            s_rt, e_rt = normal.ppf([0.02,0.98])
            # as frames
            s_frame, e_frame = device.rt_to_frame_id(s_rt), device.rt_to_frame_id(e_rt)

            profile_frames = np.arange(s_frame-1,e_frame+1) # starting with s_frame-1 for cdf interval calculation
            profile_rt_ends = device.frame_time_end(profile_frames)
            profile_rt_cdfs = normal.cdf(profile_rt_ends)
            profile_rel_intensities = np.diff(profile_rt_cdfs)

            profile_list.append(RTProfile(profile_frames[1:],profile_rel_intensities,model_params))
        return profile_list


class NeuralChromatographyApex(ChromatographyApexModel):

    def __init__(self, model_path: str, tokenizer_path: str):
        super().__init__()
        self.model_path = model_path
        self.tokenizer = tokenizer_from_json(tokenizer_path)

    def sequences_to_tokens(self, sequences_tokenized: np.array) -> np.array:
        print('tokenizing sequences...')
        tokens = self.tokenizer.texts_to_sequences(sequences_tokenized)
        tokens_padded = tf.keras.preprocessing.sequence.pad_sequences(tokens, 50, padding='post')
        return tokens_padded

    @staticmethod
    def _worker(model_path: str, tokens_padded: np.array, batched: bool = True, bs: int = 2048):
        pseudo_target = np.expand_dims(np.zeros_like(tokens_padded[:, 0]), axis=1)
        if batched:
            dataset = tf.data.Dataset.from_tensor_slices((tokens_padded, pseudo_target)).batch(bs)
        else:
            dataset = tf.data.Dataset.from_tensor_slices((tokens_padded, pseudo_target))
        model = tf.keras.models.load_model(model_path)
        print('predicting irts...')
        return_val = model.predict(dataset)
        return return_val

    def simulate(self, sample: ProteomicsExperimentSampleSlice, device: Chromatography) ->  NDArray[np.float64]:

        data = sample.peptides
        tokens = data["sequence_tokenized"].apply(lambda st: st.sequence_tokenized).to_numpy()
        print('generating tf dataset...')
        tokens_padded = self.sequences_to_tokens(tokens)

        with Pool(1) as pool:
            r = pool.starmap(self._worker, [(self.model_path, tokens_padded)])
        return r[0]


class IonSource(Device):
    def __init__(self, name:str ="IonizationDevice"):
        super().__init__(name)
        self._ionization_model = None

    @property
    def ionization_model(self):
        return self._ionization_model

    @ionization_model.setter
    def ionization_model(self, model: IonizationModel):
        self._ionization_model = model

    @abstractmethod
    def run(self, sample: ProteomicsExperimentSampleSlice):
        pass

class ElectroSpray(IonSource):
    def __init__(self, name:str ="ElectrosprayDevice"):
        super().__init__(name)

    def run(self, sample: ProteomicsExperimentSampleSlice):
        charge_profiles = self.ionization_model.simulate(sample, self)
        sample.add_simulation("simulated_charge_profile", charge_profiles)

class IonizationModel(Model):
    def __init__(self):
        pass

    @abstractmethod
    def simulate(self, sample:ProteomicsExperimentSampleSlice, device: IonSource) -> NDArray:
        pass

class RandomIonSource(IonizationModel):
    def __init__(self):
        super().__init__()
        self._charge_probabilities =  np.array([0.1, 0.5, 0.3, 0.1])
        self._allowed_charges =  np.array([1, 2, 3, 4], dtype=np.int8)

    @property
    def allowed_charges(self):
        return self._allowed_charges

    @allowed_charges.setter
    def allowed_charges(self, charges: ArrayLike):
        self._allowed_charges = np.asarray(charges, dtype=np.int8)

    @property
    def charge_probabilities(self):
        return self._charge_probabilities

    @charge_probabilities.setter
    def charge_probabilities(self, probabilities: ArrayLike):
        self._charge_probabilities = np.asarray(probabilities)

    def simulate(self, sample: ProteomicsExperimentSampleSlice, device: IonSource) -> List[ChargeProfile]:
        if self.charge_probabilities.shape != self.allowed_charges.shape:
            raise ValueError("Number of allowed charges must fit to number of probabilities")

        data = sample.peptides
        charge = np.random.choice(self.allowed_charges, data.shape[0], p=self.charge_probabilities)
        rel_intensity = np.ones_like(charge)
        charge_profiles = []
        for c,i in zip(charge, rel_intensity):
            charge_profiles.append(ChargeProfile([c],[i],model_params={"name":"RandomIonSource"}))
        return charge_profiles


class BinomialIonSource:
    def __init__(self):
        super().__init__()
        self._charged_probability = 0.5
        self._allowed_charges = np.array([1, 2, 3, 4], dtype=np.int8)

    @property
    def allowed_charges(self):
        return self._allowed_charges

    @allowed_charges.setter
    def allowed_charges(self, charges: ArrayLike):
        self._allowed_charges = np.asarray(charges, dtype=np.int8)

    @property
    def charged_probability(self):
        return self._charged_probability

    @charged_probability.setter
    def charged_probability(self, probability: float):
        self._charged_probability = probability

    def simulate(self, sample: ProteomicsExperimentSampleSlice, device: IonSource) -> List[ChargeProfile]:
        sequences = sample.peptides.sequence
        vec_get_num_protonizable_sites = np.vectorize(get_num_protonizable_sites)
        basic_aa_nums = vec_get_num_protonizable_sites(sequences)

        charge_profiles = []
        for baa_num in basic_aa_nums:
            rel_intensities = binom(baa_num,self.charged_probability).pmf(self.allowed_charges)
            charge_profiles.append(ChargeProfile(self.allowed_charges,rel_intensities,model_params={"name":"BinomialIonSource","basic_aa_num":baa_num}))
        return charge_profiles


class IonMobilitySeparation(Device):
    def __init__(self, name:str = "IonMobilityDevice"):
        super().__init__(name)
        # models
        self._apex_model = None
        self._profile_model = None

        # hardware parameter
        self._scan_intervall = 1
        self._scan_time = None
        self._scan_id_min = None
        self._scan_id_max = None
        self._buffer_gas = None

        # converters
        self._reduced_im_to_scan_converter = None
        self._scan_to_reduced_im_interval_converter = None

    @property
    def reduced_im_to_scan_converter(self):
        return self._reduced_im_to_scan_converter

    @reduced_im_to_scan_converter.setter
    def reduced_im_to_scan_converter(self, converter:callable):
        self._reduced_im_to_scan_converter = converter

    @property
    def scan_to_reduced_im_interval_converter(self):
        return self._scan_to_reduced_im_interval_converter

    @scan_to_reduced_im_interval_converter.setter
    def scan_to_reduced_im_interval_converter(self, converter:callable):
        self._scan_to_reduced_im_interval_converter = converter

    @property
    def buffer_gas(self):
        return self._buffer_gas

    @buffer_gas.setter
    def buffer_gas(self, gas: BufferGas):
        self._buffer_gas = gas

    @property
    def scan_intervall(self):
        return self._scan_intervall

    @scan_intervall.setter
    def scan_intervall(self, number:int):
        self._scan_intervall = number

    @property
    def scan_id_min(self):
        return self._scan_id_min

    @scan_id_min.setter
    def scan_id_min(self, number:int):
        self._scan_id_min = number

    @property
    def scan_id_max(self):
        return self._scan_id_max

    @scan_id_max.setter
    def scan_id_max(self, number:int):
        self._scan_id_max = number

    @property
    def scan_time(self):
        return self._scan_time

    @scan_time.setter
    def scan_time(self, microseconds:int):
        self._scan_time = microseconds

    @property
    def apex_model(self):
        return self._apex_model

    @apex_model.setter
    def apex_model(self, model: IonMobilityApexModel):
        self._apex_model = model

    @property
    def profile_model(self):
        return self._profile_model

    @profile_model.setter
    def profile_model(self, model: IonMobilityProfileModel):
        self._profile_model = model

    def scan_to_reduced_im_interval(self, scan_id: ArrayLike):
        return self.scan_to_reduced_im_interval_converter(scan_id)

    def reduced_im_to_scan(self, ion_mobility):
        return self.reduced_im_to_scan_converter(ion_mobility)

    def scan_im_middle(self, scan_id: ArrayLike):
        return np.mean(self.scan_to_reduced_im_interval(scan_id), axis = 1)

    def scan_im_lower(self, scan_id: ArrayLike):
        return self.scan_to_reduced_im_interval(scan_id)[:,0]

    def scan_im_upper(self, scan_id:ArrayLike):
        return self.scan_to_reduced_im_interval(scan_id)[:,1]

    def im_to_reduced_im(self, ion_mobility: float, p_0: float = STANDARD_PRESSURE, T_0: float = STANDARD_TEMPERATURE):
        """
        Calculate reduced ion mobility K_0
        (normalized to standard pressure p_0
        and standard temperature T_0), from
        ion mobility K.

        K_0 = K * p/p_0 * T_0/T

        [1] J. N. Dodds and E. S. Baker,
        “Ion mobility spectrometry: fundamental concepts,
        instrumentation, applications, and the road ahead,”
        Journal of the American Society for Mass Spectrometry,
        vol. 30, no. 11, pp. 2185–2195, 2019,
        doi: 10.1007/s13361-019-02288-2.

        :param ion_mobility: Ion mobility K to
            normalize to standard conditions
        :param p_0: Standard pressure (Pa).
        :param T_0: Standard temperature (K).
        """
        T = self.temperature
        p = self.pressure
        return ion_mobility*p/p_0*T_0/T

    def reduced_im_to_im(self, reduced_ion_mobility: float, p_0: float = STANDARD_PRESSURE, T_0: float = STANDARD_TEMPERATURE):
        """
        Inverse of `.im_to_reduced_im()`
        """
        T = self.temperature
        p = self.pressure
        return reduced_ion_mobility*p_0/p*T/T_0

    def ccs_to_reduced_im(self, ccs:float, mz:float, charge:int):
        # TODO Citation
        """
        Conversion of collision cross-section values (ccs)
        to reduced ion mobility according to
        Mason-Schamp equation.

        :param ccs: collision cross-section (ccs)
        :type ccs: float
        :param mz: Mass (Da) to charge ratio of peptide
        :type mz: float
        :param charge: Charge of peptide
        :type charge: int
        :return: Reduced ion mobility
        :rtype: float
        """

        T = self.temperature
        mass = mz*charge
        μ = self.buffer_gas.mass*mass/(self.buffer_gas.mass+mass)
        z = charge

        K0 = CCS_K0_CONVERSION_CONSTANT*z*1/(np.sqrt(μ*T)*ccs)
        return K0

    def reduced_im_to_ccs(self, reduced_ion_mobility:float, mz:float, charge:int):
        """
        Conversion of reduced ion mobility
        to collision cross-section values (ccs)
        according to Mason-Schamp equation.

        :param reduced_ion_mobility: reduced ion mobility K0
        :type reduced_ion_mobility: float
        :param mz: Mass (Da) to charge ratio of peptide
        :type mz: float
        :param charge: Charge of peptide
        :type charge: int
        :return: Collision cross-section (ccs)
        :rtype: float
        """

        T = self.temperature
        mass = mz*charge
        μ = self.buffer_gas.mass*mass/(self.buffer_gas.mass+mass)
        z = charge
        K0 = reduced_ion_mobility

        ccs = CCS_K0_CONVERSION_CONSTANT*z*1/(np.sqrt(μ*T)*K0)
        return ccs

    @abstractmethod
    def run(self, sample: ProteomicsExperimentSampleSlice):
        pass


class TrappedIon(IonMobilitySeparation):

    def __init__(self, name:str = "TrappedIonMobilitySeparation"):
        super().__init__()
        self._scan_id_min = 1
        self._scan_id_max = 918

    def run(self, sample: ProteomicsExperimentSampleSlice):
        # scan apex simulation
        one_over_k0, scan_apex = self._apex_model.simulate(sample, self)
        # in irt and rt
        sample.add_simulation("simulated_scan_apex", scan_apex)
        sample.add_simulation("simulated_k0", one_over_k0)
        # scan profile simulation
        scan_profile = self._profile_model.simulate(sample, self)
        sample.add_simulation("simulated_scan_profile", scan_profile)


class IonMobilityApexModel(Model):
    def __init__(self):
        pass

    @abstractmethod
    def simulate(self, sample: ProteomicsExperimentSampleSlice, device: IonMobilitySeparation) -> NDArray[np.float64]:
        return super().simulate(sample, device)


class IonMobilityProfileModel(Model):
    def __init__(self):
        pass

    @abstractmethod
    def simulate(self, sample: ProteomicsExperimentSampleSlice, device: IonMobilitySeparation) -> NDArray:
        return super().simulate(sample, device)


class NeuralIonMobilityApex(IonMobilityApexModel):

    def __init__(self, model_path: str, tokenizer_path: str):
        super().__init__()
        self.model_path = model_path
        self.tokenizer = tokenizer_from_json(tokenizer_path)

    def sequences_to_tokens(self, sequences_tokenized: np.array) -> np.array:
        print('tokenizing sequences...')
        tokens = self.tokenizer.texts_to_sequences(sequences_tokenized)
        tokens_padded = tf.keras.preprocessing.sequence.pad_sequences(tokens, 50, padding='post')
        return tokens_padded

    @staticmethod
    def _worker(model_path: str, tokens_padded: np.array, mz: np.array, charges: np.array, batched: bool = True, bs: int = 2048):

        mz = np.expand_dims(mz, 1)
        c = tf.one_hot(charges - 1, depth=4)
        pseudo_target = np.expand_dims(np.zeros_like(tokens_padded[:, 0]), axis=1)
        if batched:
            dataset = tf.data.Dataset.from_tensor_slices(((mz, c, tokens_padded), pseudo_target)).batch(bs)
        else:
            dataset = tf.data.Dataset.from_tensor_slices(((mz, c, tokens_padded), pseudo_target))

        print('predicting mobilities...')
        model = tf.keras.models.load_model(model_path)
        ccs, _ = model.predict(dataset)
        return ccs

    def simulate(self, sample: ProteomicsExperimentSampleSlice, device: IonMobilitySeparation) -> Tuple[NDArray]:
        data = sample.ions.merge(sample.peptides.loc[:,["pep_id","sequence_tokenized"]],on="pep_id",validate="many_to_one")
        tokens = data.sequence_tokenized.apply(lambda st: st.sequence_tokenized).to_numpy()
        tokens_padded = self.sequences_to_tokens(tokens)
        mz = data['mz'].values
        charges = data["charge"].values
        with Pool(1) as pool:
            r = pool.starmap(self._worker, [(self.model_path, tokens_padded, mz, charges)])
            ccs = r[0]
        K0s = device.ccs_to_reduced_im(np.squeeze(ccs), mz, data['charge'].values)
        scans = device.reduced_im_to_scan(K0s)
        return K0s,scans


class NormalIonMobilityProfileModel(IonMobilityProfileModel):
    def __init__(self):
        super().__init__()

    def simulate(self, sample: ProteomicsExperimentSampleSlice, device: IonMobilitySeparation) -> List[ScanProfile]:
        mus = 1/sample.ions["simulated_k0"].values #1 over k0
        sigmas = gamma(a=3.703,scale=1/79.614).rvs(mus.size)/6
        profile_list = []
        for μ,σ in zip(mus,sigmas):

            model_params = { "sigma":σ,
                             "mu":μ,
                             "name":"NORMAL"
                            }

            normal = norm(loc=μ, scale=σ)
            # start and end value (k0)
            s_one_over_k0, e_one_over_k0 = normal.ppf([0.01,0.99])
            # as scan ids, remember first scans elutes largest ions
            e_scan, s_scan = device.reduced_im_to_scan(1/s_one_over_k0), device.reduced_im_to_scan(1/e_one_over_k0)

            profile_scans = np.arange(s_scan-1,e_scan+1) # starting s_scan-1 is necessary here to include its end value for cdf interval
            profile_end_im = 1/device.scan_im_upper(profile_scans)
            profile_end_cdfs = 1-normal.cdf(profile_end_im)
            profile_rel_intensities = np.diff(profile_end_cdfs)

            profile_list.append(ScanProfile(profile_scans[1:],profile_rel_intensities,model_params))
        return profile_list


class MzSeparation(Device):
    def __init__(self, name:str = "MassSpectrometer"):
        super().__init__(name)
        self._model = None
        self._resolution = 3

    @property
    def resolution(self):
        return self._resolution

    @resolution.setter
    def resolution(self, res: int):
        self._resolution = res

    @property
    def model(self):
        return self._model

    @model.setter
    def model(self, model: MzSeparationModel):
        self._model = model

    @abstractmethod
    def run(self, sample: ProteomicsExperimentSampleSlice):
        pass


class TOF(MzSeparation):
    def __init__(self, name:str = "TimeOfFlightMassSpectrometer"):
        super().__init__(name)

    def run(self, sample: ProteomicsExperimentSampleSlice):
        spectra = self.model.simulate(sample, self)
        sample.add_simulation("simulated_mz_spectrum", spectra)


class MzSeparationModel(Model):
    def __init__(self):
        self._default_abundance = 1e4

    @property
    def default_abundance(self):
        return self._default_abundance

    @default_abundance.setter
    def default_abundance(self, abundance:int):
        self._default_abundance = abundance

    @abstractmethod
    def simulate(self, sample: ProteomicsExperimentSampleSlice, device: MzSeparation):
        return super().simulate(sample, device)


class AveragineModel(MzSeparationModel):
    def __init__(self):
        super().__init__()

    def simulate(self, sample: ProteomicsExperimentSampleSlice, device: MzSeparation):

        avg = AveragineGenerator()
        # right join here to keep order of keys in right df
        joined_df = pd.merge(sample.peptides[["pep_id","mass_theoretical"]],
                        sample.ions[["pep_id","charge"]],
                        how = "right",
                        on = "pep_id")
        masses = joined_df["mass_theoretical"].values
        charges = joined_df["charge"].values
        spectra = []
        for (m, c) in zip(masses, charges):
            spectrum = avg.generate_spectrum(m,c,amp = self.default_abundance, centroided=False)
            spectra.append(spectrum)
        return spectra



import tensorflow as tf
import numpy as np
import tensorflow_probability as tfp

tfd = tfp.distributions


class GaussianMixtureModel(tf.Module):

    def __init__(self, num_components: int, data_dim: int, prior_stddevs=None, data=None,
                 init_means=None, init_stds=None):
        """
        Initialize the Gaussian Mixture Model.

        Parameters:
        - num_components: Number of Gaussian components.
        - data_dim: Dimensionality of the data.
        - prior_stdevs (optional): Prior knowledge about cluster extensions (standard deviations).
        - lambda_scale: Regularization strength for the scales.
        - data (optional): If provided and no init_means is given, initialize the component means by randomly selecting from this data.
        - init_means (optional): Explicit initial means for the components.
        - init_scales (optional): Explicit initial scales (variances) for the components.
        """

        # Initialize the locations of the GMM components
        super().__init__()

        if init_means is not None:
            assert init_means.shape == (num_components,
                                        data_dim), f"init_means should have shape [num_components, data_dim], but got {init_means.shape}"
            init_locs = tf.convert_to_tensor(init_means, dtype=tf.float32)

        elif data is not None:
            indices = np.random.choice(data.shape[0], size=num_components, replace=True,)
            init_locs = tf.convert_to_tensor(data[indices], dtype=tf.float32)

        else:
            init_locs = tf.random.normal([num_components, data_dim])

        self.locs = tf.Variable(init_locs, name="locs")

        if init_stds is not None:
            init_stds_vals = tf.repeat(init_stds, num_components, axis=0)
        else:
            init_stds_default = [[3, 0.01, 0.01]]
            init_stds_vals = tf.repeat(init_stds_default, num_components, axis=0)

        self.scales = tf.Variable(tf.math.log(init_stds_vals), name="scales")

        # Initialize the weights of the GMM components
        self.weights = tf.Variable(tf.ones([num_components]), name="weights")

        # Set the prior scales and regularization strength
        if prior_stddevs is not None:
            init_prior_stds = tf.repeat(prior_stddevs, num_components, axis=0)
            self.prior_stddevs = init_prior_stds
        else:
            self.prior_stddevs = None

    def __call__(self, data):
        """
        Calculate the log likelihood of the data given the current state of the GMM.
        """

        # Constructing the multivariate normal distribution with diagonal covariance
        components_distribution = tfd.Independent(tfd.Normal(loc=self.locs, scale=tf.math.exp(self.scales)),
                                                  reinterpreted_batch_ndims=1)

        gmm = tfd.MixtureSameFamily(
            mixture_distribution=tfd.Categorical(logits=tf.nn.log_softmax(self.weights)),
            components_distribution=components_distribution)

        return gmm.log_prob(data)

    def fit(self, data, weights=None, num_steps=200, learning_rate=0.05, lambda_scale=0.01, verbose=True):
        """
        Fit the Gaussian Mixture Model to the data.

        Parameters:
        - data: Input data of shape [n_samples, n_features].
        - weights (optional): Weights for each sample.
        - num_steps: Number of optimization steps.
        - learning_rate: Learning rate for the optimizer.
        """
        if weights is None:
            weights = tf.ones(len(data))

        optimizer = tf.optimizers.Adam(learning_rate=learning_rate)

        @tf.function
        def train_step():
            with tf.GradientTape() as tape:
                log_likelihood = self.__call__(data)
                loss = -tf.reduce_sum(log_likelihood * weights)  # Weighted negative log likelihood

                # Add regularization based on prior scales if provided

                if self.prior_stddevs is not None:
                    scale_diff = tf.math.exp(self.scales) - self.prior_stddevs
                    reg_loss = lambda_scale * tf.reduce_sum(scale_diff * scale_diff)
                    loss += reg_loss

            gradients = tape.gradient(loss, [self.locs, self.scales, self.weights])
            optimizer.apply_gradients(zip(gradients, [self.locs, self.scales, self.weights]))
            return loss

        for step in range(num_steps):
            loss = train_step()
            if step % 100 == 0 and verbose:
                tf.print("step:", step, "log-loss:", loss)

    def _mixture(self):
        """
        Creates a Gaussian Mixture Model from the current parameters (weights, means, and covariances).
        """
        return tfd.MixtureSameFamily(
            mixture_distribution=tfd.Categorical(logits=tf.nn.log_softmax(self.weights)),
            components_distribution=tfd.Independent(tfd.Normal(loc=self.locs, scale=tf.math.exp(self.scales)),
                                                    reinterpreted_batch_ndims=1))

    @property
    def variances(self):
        """Returns the actual variances (squared scales) of the Gaussian components."""
        return tf.math.exp(2 * self.scales)

    @property
    def stddevs(self):
        """Returns the actual standard deviations (scales) of the Gaussian components."""
        return tf.math.exp(self.scales)

    def predict_proba(self, data):
        gmm = self._mixture()

        # Calculate the log probabilities for each data point for each component
        log_probs = gmm.components_distribution.log_prob(tf.transpose(data[..., tf.newaxis], [0, 2, 1]))

        # Convert log probabilities to unnormalized probabilities
        unnormalized_probs = tf.exp(log_probs)

        # Normalize the probabilities
        probs_sum = tf.reduce_sum(unnormalized_probs, axis=-1, keepdims=True)
        normalized_probs = unnormalized_probs / probs_sum

        return normalized_probs.numpy()

    def predict(self, data):
        """Get the cluster ids under the current mixture model"""
        return np.argmax(self.predict_proba(data), axis=1)

    def sample(self, n_samples=1):
        gmm = self._mixture()

        # Sample from the Gaussian Mixture Model
        samples = gmm.sample(n_samples)

        return samples


import tensorflow as tf
import numpy as np
import warnings


class CosimHasher:
    def __init__(self, target_vector_length, trials: int = 32, len_trial: int = 20, seed: int = 42):

        assert 0 < trials, f'trials variable needs to be greater then 1, was: {trials}'
        assert 0 < len_trial, f'length trial variable needs to be greater then 1, was: {trials}'

        # check
        if 0 < len_trial <= 32:
            self.V = tf.constant(
                np.expand_dims(np.array([np.power(2, i) for i in range(len_trial)]).astype(np.int32), 1))

        elif 32 < len_trial <= 64:
            warnings.warn(f"\nnum bits to hash set to: {len_trial}.\n" +
                          f"using int64 which might slow down computation significantly.")
            self.V = tf.constant(
                np.expand_dims(np.array([np.power(2, i) for i in range(len_trial)]).astype(np.int64), 1))
        else:
            raise ValueError(f"bit number per hash cannot be greater then 64 or smaller 1, was: {len_trial}.")

        self.trials = trials
        self.len_trial = len_trial
        self.seed = seed
        self.target_vector_length = target_vector_length

        np.random.seed(seed)
        size = (len_trial * self.trials, target_vector_length)
        X = np.random.normal(0, 1, size=size).astype(np.float32)
        self.hash_tensor = tf.transpose(tf.constant(X))

    def __repr__(self) -> str:
        return f"CosimHasher(trials={self.trials}, len_trial={self.len_trial}, seed={self.seed}, " \
               f"target_vector_length={self.target_vector_length})"

    def calculate_keys(self, W: tf.Tensor) -> tf.Tensor:

        S = (tf.sign(W @ self.hash_tensor) + 1) / 2

        if self.len_trial <= 32:
            # reshape into window, num_hashes, len_single_hash
            S = tf.cast(tf.reshape(S, shape=(S.shape[0], self.trials, self.len_trial)), tf.int32)

            # calculate int key from binary by base-transform
            H = tf.squeeze(S @ self.V)
            return H
        else:
            # reshape into window, num_hashes, len_single_hash
            S = tf.cast(tf.reshape(S, shape=(S.shape[0], self.trials, self.len_trial)), tf.int64)

            # calculate int key from binary by base-transform
            H = tf.squeeze(S @ self.V)
            return H


class TimsHasher(CosimHasher):
    """
    Class to create hash keys from a given set of weights.

    Args:

        trials (int): number of trials to use for random projection.
        len_trial (int): length of each trial.
        seed (int): seed for random projection.
        resolution (int): resolution of the random projection.
        num_dalton (int): number of dalton to use for random projection.
    """
    def __init__(self, trials: int = 32, len_trial: int = 20, seed: int = 5671, resolution: int = 1, num_dalton: int = 10):
        res_factor = 10 ** resolution
        target_vector_length = num_dalton * res_factor + 1
        super().__init__(target_vector_length, trials, len_trial, seed)
        self.resolution = resolution
        self.num_dalton = num_dalton

    def __repr__(self) -> str:
        return f"TimsHasher(trials={self.trials}, len_trial={self.len_trial}, seed={self.seed}, " \
               f"resolution={self.resolution}, num_dalton={self.num_dalton})"



from numpy.typing import NDArray
from typing import List

import numpy as np

import pandas as pd

import tensorflow as tf
from imspy.simulation.utility import sequence_to_numpy
from dlomix.constants import ALPHABET_UNMOD

from tensorflow.keras.layers.experimental import preprocessing

from dlomix.reports.postprocessing import (reshape_flat, reshape_dims,
                                           normalize_base_peak, mask_outofcharge, mask_outofrange)


def seq_to_index(seq: str) -> List[int]:
    """Convert a sequence to a list of indices into the alphabet.

    Args:
        seq: A string representing a sequence of amino acids.

    Returns:
        A list of integers, each representing an index into the alphabet.
    """
    return [ALPHABET_UNMOD[s] for s in seq]


# Your existing code for data preparation, with modifications to name the inputs
def generate_prosit_intensity_prediction_dataset(
        sequences: List[str],
        charges: NDArray,
        collision_energies: NDArray | None = None):

    # set default for unprovided collision_energies
    if collision_energies is None:
        collision_energies = np.expand_dims(np.repeat([0.35], len(charges)), 1)

    # check for 1D collision_energies, need to be expanded to 2D
    elif len(collision_energies.shape) == 1:
        collision_energies = np.expand_dims(collision_energies, 1)

    # Create a string lookup layer to convert sequences to indices, and one-hot encode charges
    string_lookup = preprocessing.StringLookup(
        vocabulary=list(ALPHABET_UNMOD.keys())
    )

    charges = tf.one_hot(charges - 1, depth=6)
    sequences = tf.cast(string_lookup([sequence_to_numpy(s) for s in sequences]), dtype=tf.int32)

    # Create a dataset that yields batches in the format expected by the model??
    dataset = tf.data.Dataset.from_tensor_slices((
        {"peptides_in": sequences,
         "precursor_charge_in": charges,
         "collision_energy_in": tf.cast(collision_energies, dtype=tf.float32)}
    ))

    return dataset


def unpack_dict(features):
    """Unpack the dictionary of features into the expected inputs for the model.

    Args:
        features: A dictionary of features, with keys 'peptides_in', 'precursor_charge_in', and 'collision_energy_in'

    Returns:
        A tuple of the expected inputs for the model: (peptides, precursor_charge, collision_energy)
    """
    return features['peptides_in'], features['precursor_charge_in'], features['collision_energy_in']


def post_process_predicted_fragment_spectra(data_pred: pd.DataFrame) -> NDArray:
    """
    post process the predicted fragment intensities
    Args:
        data_pred: dataframe containing the predicted fragment intensities

    Returns:
        numpy array of fragment intensities
    """
    # get sequence length for masking out of sequence
    sequence_lengths = data_pred["sequence_length"]

    # get data
    intensities = np.stack(data_pred["intensity_raw"].to_numpy()).astype(np.float32)
    # set negative intensity values to 0
    intensities[intensities < 0] = 0
    intensities = reshape_dims(intensities)

    # mask out of sequence and out of charge
    intensities = mask_outofrange(intensities, sequence_lengths)
    intensities = mask_outofcharge(intensities, data_pred.charge)
    intensities = reshape_flat(intensities)

    # save indices of -1.0 values, will be altered by intensity normalization
    m_idx = intensities == -1.0
    # normalize to base peak
    intensities = normalize_base_peak(intensities)
    intensities[m_idx] = -1.0

    return intensities


def to_prosit_tensor(sequences: List) -> tf.Tensor:
    """
    translate a list of fixed length numpy arrays into a tensorflow tensor
    Args:
        sequences: list of numpy arrays, representing peptide sequences

    Returns:
        tensorflow tensor
    """
    return tf.convert_to_tensor(sequences, dtype=tf.string)



from typing import List, Tuple, Union

import imspy_connector

from imspy.data.spectrum import MzSpectrum

ims = imspy_connector.py_peptide


class PeptideProductIonSeriesCollection:
    def __init__(self, series: List['PeptideProductIonSeries']):
        """Create a new product ion series collection.

        Args:
            series: The product ion series.
        """
        self.__ptr = ims.PyPeptideProductIonSeriesCollection([s.get_ptr() for s in series])

    @property
    def series(self) -> List['PeptideProductIonSeries']:
        return [PeptideProductIonSeries.from_py_ptr(series) for series in self.__ptr.series]

    def find_series(self, charge: int) -> Union[None, 'PeptideProductIonSeries']:
        """Find the product ion series with the given charge.

        Args:
            charge: The charge of the product ion series.

        Returns:
            The product ion series with the given charge, or None if not found.
        """
        series = self.__ptr.find_ion_series(charge)
        return PeptideProductIonSeries.from_py_ptr(series) if series else None

    def get_ptr(self):
        return self.__ptr

    def to_json(self) -> str:
        return self.__ptr.to_json()

    @classmethod
    def from_py_ptr(cls, collection: ims.PyPeptideProductIonSeriesCollection):
        instance = cls.__new__(cls)
        instance.__ptr = collection
        return instance

    def __repr__(self):
        return f"PeptideProductIonSeriesCollection(series={self.series})"

    # mass_tolerance: f64, abundance_threshold: f64, max_result: i32, intensity_min: f64
    def generate_isotopic_spectrum(
            self,
            mass_tolerance: float = 1e-3,
            abundance_threshold: float = 1e-6,
            max_result: int = 2000,
            intensity_min: float = 1e-6
    ) -> MzSpectrum:
        """Calculate the isotope distribution of the product ion series collection.

        Args:
            mass_tolerance: The mass tolerance for the isotope distribution calculation.
            abundance_threshold: The abundance threshold for the isotope distribution calculation.
            max_result: The maximum number of results to return.
            intensity_min: The minimum intensity of the isotope distribution.

        Returns:
            The isotope distribution of the product ion series collection.
        """
        py_spec = self.__ptr.generate_isotopic_spectrum(mass_tolerance, abundance_threshold, max_result, intensity_min)
        return MzSpectrum.from_py_mz_spectrum(py_spec)


class PeptideProductIonSeries:
    def __init__(
            self,
            charge: int,
            n_ions: List['PeptideProductIon'],
            c_ions: List['PeptideProductIon']
    ):
        """Create a new product ion series.

        Args:
            charge: The charge of the product ions.
            n_ions: The N-terminal product ions.
            c_ions: The C-terminal product ions.
        """
        self.__ptr = ims.PyPeptideProductIonSeries(
            charge=charge,
            n_ions=[ion.get_ptr() for ion in n_ions],
            c_ions=[ion.get_ptr() for ion in c_ions]
        )

    @property
    def n_ions(self) -> List['PeptideProductIon']:
        return [PeptideProductIon.from_py_ptr(ion) for ion in self.__ptr.n_ions]

    @property
    def c_ions(self) -> List['PeptideProductIon']:
        return [PeptideProductIon.from_py_ptr(ion) for ion in self.__ptr.c_ions]

    @property
    def charge(self) -> int:
        return self.__ptr.charge

    def get_ptr(self):
        return self.__ptr

    def to_json(self) -> str:
        return self.__ptr.to_json()

    @classmethod
    def from_py_ptr(cls, series: ims.PyPeptideProductIonSeries):
        instance = cls.__new__(cls)
        instance.__ptr = series
        return instance

    def __repr__(self):
        return (f"PeptideProductIonSeries(charge={self.charge}, n_ions={self.n_ions}, "
                f"c_ions={self.c_ions})")


class PeptideProductIon:
    def __init__(self, kind: str, sequence: str, charge: int = 1, intensity: float = 1.0):
        """Create a new product ion.

        Args:
            kind: The kind of product ion, must be one of 'a', 'b', 'c', 'x', 'y', 'z'.
            sequence: The sequence of the product ion.
            charge: The charge of the product ion.
            intensity: The intensity of the product ion.
        """
        assert kind in ['a', 'b', 'c', 'x', 'y', 'z'], (f"Invalid kind: {kind}, "
                                                        f"must be one of 'a', 'b', 'c', 'x', 'y', 'z'")
        self.__ptr = ims.PyPeptideProductIon(
            kind=kind,
            sequence=sequence,
            charge=charge,
            intensity=intensity
        )

    @property
    def kind(self) -> str:
        return self.__ptr.kind

    @property
    def sequence(self) -> str:
        return self.__ptr.sequence

    @property
    def charge(self) -> int:
        return self.__ptr.charge

    @property
    def intensity(self) -> float:
        return self.__ptr.intensity

    @property
    def mono_isotopic_mass(self) -> float:
        return self.__ptr.mono_isotopic_mass()

    @property
    def mz(self) -> float:
        return self.__ptr.mz

    def atomic_composition(self):
        return self.__ptr.atomic_composition()

    def to_json(self) -> str:
        return self.__ptr.to_json()

    def get_ptr(self):
        return self.__ptr

    @classmethod
    def from_py_ptr(cls, product_ion: ims.PyPeptideProductIon):
        instance = cls.__new__(cls)
        instance.__ptr = product_ion
        return instance

    def isotope_distribution(
            self,
            mass_tolerance: float = 1e-3,
            abundance_threshold: float = 1e-8,
            max_result: int = 200,
            intensity_min: float = 1e-4
    ) -> List[Tuple[float, float]]:
        """Calculate the isotope distribution of the product ion.

        Args:
            mass_tolerance: The mass tolerance for the isotope distribution calculation.
            abundance_threshold: The abundance threshold for the isotope distribution calculation.
            max_result: The maximum number of results to return.
            intensity_min: The minimum intensity of the isotope distribution.

        Returns:
            The isotope distribution of the product ion.
        """
        return self.__ptr.isotope_distribution(mass_tolerance, abundance_threshold, max_result, intensity_min)

    def __repr__(self):
        return (f"ProductIon(kind={self.kind}, sequence={self.sequence}, charge={self.charge}, mz={self.mz}, "
                f" intensity={self.intensity})")


class PeptideSequence:
    def __init__(self, sequence: str):
        """Create a new peptide sequence.

        Args:
            sequence: The sequence of the peptide.
        """
        self.__ptr = ims.PyPeptideSequence(sequence)

    @property
    def sequence(self) -> str:
        return self.__ptr.sequence

    @property
    def mono_isotopic_mass(self) -> float:
        return self.__ptr.mono_isotopic_mass

    @property
    def atomic_composition(self):
        return self.__ptr.atomic_composition()

    @property
    def amino_acid_count(self) -> int:
        return self.__ptr.amino_acid_count()

    def to_tokens(self, group_modifications: bool = True) -> List[str]:
        return self.__ptr.to_tokens(group_modifications)

    def to_sage_representation(self) -> Tuple[str, List[float]]:
        return self.__ptr.to_sage_representation()

    def to_json(self) -> str:
        return self.__ptr.to_json()

    def get_ptr(self):
        return self.__ptr

    def calculate_product_ion_series(
            self,
            charge: int = 1,
            fragment_type:
            str = 'b'
    ) -> Tuple[List[PeptideProductIon], List[PeptideProductIon]]:
        """Calculate the b and y product ion series of the peptide sequence.

        Args:
            charge: The charge of the product ions.
            fragment_type: The type of the product ions, must be one of 'a', 'b', 'c', 'x', 'y', 'z'.

        Returns:
            The b and y product ion series of the peptide sequence.
        """
        fragment_type = fragment_type.lower()
        assert fragment_type in ['a', 'b', 'c', 'x', 'y', 'z'], (f"Invalid fragment type: {fragment_type}, "
                                                                 f"must be one of 'a', 'b', 'c', 'x', 'y', 'z'")

        n_ions, c_ions = self.__ptr.calculate_product_ion_series(charge, fragment_type)

        return ([PeptideProductIon.from_py_ptr(ion) for ion in n_ions],
                [PeptideProductIon.from_py_ptr(ion) for ion in c_ions][::-1])

    def calculate_mono_isotopic_product_ion_spectrum(
            self,
            charge:
            int = 1,
            fragment_type: str = 'b'
    ) -> MzSpectrum:
        """Calculate the mono-isotopic product ion spectrum of the peptide sequence.
        
        Args:
            charge: The charge of the product ions.
            fragment_type: The type of the product ions, must be one of 'a', 'b', 'c', 'x', 'y', 'z'.
            
        Returns:
            The mono-isotopic product ion spectrum of the peptide sequence.
        """
        fragment_type = fragment_type.lower()
        assert fragment_type in ['a', 'b', 'c', 'x', 'y', 'z'], (f"Invalid fragment type: {fragment_type}, "
                                                                 f"must be one of 'a', 'b', 'c', 'x', 'y', 'z'")

        py_spec = self.__ptr.calculate_mono_isotopic_product_ion_spectrum(charge, fragment_type)
        return MzSpectrum.from_py_mz_spectrum(py_spec)

    def associate_fragment_ion_series_with_prosit_intensities(
            self, flat_intensities: List[float],
            charge: int,
            fragment_type: str = "b",
            normalize: bool = True,
            half_charge_one: bool = True
    ) -> PeptideProductIonSeriesCollection:
        """Associate the peptide sequence with predicted intensities from Prosit intensity prediction.

        Args:
            flat_intensities: The flat intensities.
            # TODO: check how charge should be handled
            charge: The charge of the product ions.
            fragment_type: The type of the product ions, must be one of 'a', 'b', 'c', 'x', 'y', 'z'.
            normalize: Whether to normalize the intensities.
            half_charge_one: Whether to use half charge one.

        Returns:
            The b and y product ion series of the peptide sequence.
        """
        fragment_type = fragment_type.lower()
        assert fragment_type in ['a', 'b', 'c', 'x', 'y', 'z'], (f"Invalid fragment type: {fragment_type}, "
                                                                 f"must be one of 'a', 'b', 'c', 'x', 'y', 'z'")

        result = self.__ptr.associate_with_predicted_intensities(flat_intensities,
                                                                 charge, fragment_type, normalize, half_charge_one)

        return PeptideProductIonSeriesCollection.from_py_ptr(result)

    @classmethod
    def fom_py_ptr(cls, seq: ims.PyPeptideSequence):
        instance = cls.__new__(cls)
        instance.__ptr = seq
        return instance

    def __repr__(self):
        return f"PeptideSequence(sequence={self.sequence}, mono_isotopic_mass={self.mono_isotopic_mass})"


class PeptideIon:
    def __init__(self, sequence: str, charge: int, intensity: float):
        """Create a new peptide ion.

        Args:
            sequence: The sequence of the peptide ion.
            charge: The charge of the peptide ion.
            intensity: The intensity of the peptide ion.
        """
        self.__ptr = ims.PyPeptideIon(sequence, charge, intensity)

    @property
    def sequence(self) -> PeptideSequence:
        return PeptideSequence.fom_py_ptr(self.__ptr.sequence)

    @property
    def charge(self) -> int:
        return self.__ptr.charge

    @property
    def intensity(self) -> float:
        return self.__ptr.intensity

    @property
    def mz(self) -> float:
        return self.__ptr.mz

    @property
    def atomic_composition(self):
        return self.__ptr.atomic_composition()

    def calculate_isotopic_spectrum(
            self,
            mass_tolerance: float = 1e-3,
            abundance_threshold: float = 1e-8,
            max_result: int = 200,
            intensity_min: float = 1e-4
    ) -> MzSpectrum:
        py_spec = self.__ptr.calculate_isotopic_spectrum(mass_tolerance, abundance_threshold, max_result, intensity_min)
        return MzSpectrum.from_py_mz_spectrum(py_spec)

    def get_ptr(self):
        return self.__ptr

    def to_json(self) -> str:
        return self.__ptr.to_json()

    @classmethod
    def from_py_ptr(cls, ion: ims.PyPeptideIon):
        instance = cls.__new__(cls)
        instance.__ptr = ion
        return instance

    def __repr__(self):
        return f"PeptideIon(sequence={self.sequence}, charge={self.charge}, mz={self.mz}, intensity={self.intensity})"



import re
from typing import List

from numpy.typing import NDArray
import pandas as pd
import numpy as np
import tensorflow as tf
from abc import ABC, abstractmethod
from tqdm import tqdm

from imspy.algorithm.utility import get_model_path
from imspy.algorithm.intensity.utility import (generate_prosit_intensity_prediction_dataset, unpack_dict,
                                               post_process_predicted_fragment_spectra, reshape_dims)
from imspy.data.peptide import PeptideProductIonSeriesCollection, PeptideSequence

from imspy.simulation.utility import flatten_prosit_array


def remove_unimod_annotation(sequence: str) -> str:
    """Remove [UNIMOD:N] annotations from the sequence."""
    pattern = r'\[UNIMOD:\d+\]'
    return re.sub(pattern, '', sequence)


def load_prosit_2023_timsTOF_predictor():
    """ Get a pretrained deep predictor model
    This model was downloaded from ZENODO: https://zenodo.org/records/8211811
    PAPER : https://doi.org/10.1101/2023.07.17.549401
    Returns:
        The pretrained deep predictor model
    """
    return tf.saved_model.load(get_model_path('Prosit2023TimsTOFPredictor'))


class IonIntensityPredictor(ABC):
    """
    ABSTRACT INTERFACE for simulation of ion-mobility apex value
    """

    def __init__(self):
        pass

    @abstractmethod
    def simulate_ion_intensities(self, sequences: list[str], charges: list[int], collision_energies) -> NDArray:
        pass

    @abstractmethod
    def simulate_ion_intensities_pandas(self, data: pd.DataFrame) -> pd.DataFrame:
        pass


class Prosit2023TimsTofWrapper(IonIntensityPredictor):
    """
    Wrapper for the Prosit 2023 TIMS-TOF predictor
    """
    def __init__(self, verbose: bool = True, model_name: str = 'deep_ion_intensity_predictor'):
        super().__init__()

        self.verbose = verbose
        self.model_name = model_name
        self.model = load_prosit_2023_timsTOF_predictor()

    def simulate_ion_intensities_pandas_batched(
            self,
            data: pd.DataFrame,
            batch_size_tf_ds: int = 1024,
            batch_size: int = int(4e5),
            divide_collision_energy_by: float = 1e2,
    ) -> pd.DataFrame:

        tables = []

        batch_counter = 0
        for batch_indices in tqdm(np.array_split(data.index, np.ceil(len(data) / batch_size)),
                                  total=int(np.ceil(len(data) / batch_size)),
                                  desc='Simulating intensities', ncols=100, disable=not self.verbose):

            batch = data.loc[batch_indices].reset_index(drop=True)
            data_pred = self.simulate_ion_intensities_pandas(batch, batch_size=batch_size_tf_ds,
                                                             divide_collision_energy_by=divide_collision_energy_by)

            # TODO: Save the data_pred to a database per batch, so that we don't have to keep everything in memory
            """
            if batch_counter == 0:
                handle.create_table('intensity_predictions', data_pred)
            """

            tables.append(data_pred)
            batch_counter += 1

        return pd.concat(tables)

    def simulate_ion_intensities_pandas(self, data: pd.DataFrame, batch_size: int = 512,
                                        divide_collision_energy_by: float = 1e2, verbose: bool = False) -> pd.DataFrame:

        if verbose:
            print("Generating Prosit compatible input data...")

        data['sequence_unmod'] = data.apply(lambda r: remove_unimod_annotation(r.sequence), axis=1)
        data['collision_energy'] = data.apply(lambda r: r.collision_energy / divide_collision_energy_by, axis=1)
        data['sequence_length'] = data.apply(lambda r: len(r.sequence_unmod), axis=1)

        tf_ds = (generate_prosit_intensity_prediction_dataset(
            data.sequence_unmod,
            data.charge,
            np.expand_dims(data.collision_energy, 1)).batch(batch_size))

        # Map the unpacking function over the dataset
        ds_unpacked = tf_ds.map(unpack_dict)

        intensity_predictions = []

        # Iterate over the dataset and call the model with unpacked inputs
        for peptides_in, precursor_charge_in, collision_energy_in in tqdm(ds_unpacked, desc='Predicting intensities',
                                                                          total=len(data) // batch_size + 1, ncols=100,
                                                                          disable=not verbose):
            model_input = [peptides_in, precursor_charge_in, collision_energy_in]
            model_output = self.model(model_input).numpy()
            intensity_predictions.append(model_output)

        I_pred = list(np.vstack(intensity_predictions))
        data['intensity_raw'] = I_pred
        I_pred = np.squeeze(reshape_dims(post_process_predicted_fragment_spectra(data)))

        data['intensity'] = list(I_pred)

        return data

    def simulate_ion_intensities(
            self,
            sequences: List[str],
            charges: List[int],
            collision_energies: List[float],
            divide_collision_energy_by: float = 1e2,
            batch_size: int = 512,
    ) -> List[PeptideProductIonSeriesCollection]:
        sequences_unmod = [remove_unimod_annotation(s) for s in sequences]
        sequence_length = [len(s) for s in sequences_unmod]
        collision_energies_norm = [ce / divide_collision_energy_by for ce in collision_energies]

        tf_ds = generate_prosit_intensity_prediction_dataset(
            sequences_unmod,
            charges,
            np.expand_dims(collision_energies_norm, 1)).batch(batch_size)

        ds_unpacked = tf_ds.map(unpack_dict)

        intensity_predictions = []
        for peptides_in, precursor_charge_in, collision_energy_in in tqdm(ds_unpacked, desc='Predicting intensities',
                                                                          total=len(sequences) // batch_size + 1, ncols=100,
                                                                          disable=not self.verbose):
            model_input = [peptides_in, precursor_charge_in, collision_energy_in]
            model_output = self.model(model_input).numpy()
            intensity_predictions.append(model_output)

        I_pred = list(np.vstack(intensity_predictions))
        I_pred = np.squeeze(reshape_dims(post_process_predicted_fragment_spectra(pd.DataFrame({
            'sequence': sequences,
            'charge': charges,
            'collision_energy': collision_energies,
            'sequence_length': sequence_length,
            'intensity_raw': I_pred,
        }))))

        intensities = np.vstack([flatten_prosit_array(r) for r in I_pred])
        peptide_sequences = [PeptideSequence(s) for s in sequences]
        ion_collections = []

        for peptide, charge, intensity in zip(peptide_sequences, charges, intensities):
            series = peptide.associate_fragment_ion_series_with_prosit_intensities(
                intensity,
                charge
            )
            ion_collections.append(series)

        return ion_collections



import numpy as np
import pandas as pd
import sqlite3

from typing import Dict, List

from numpy.typing import NDArray
import opentims_bruker_bridge as obb

from abc import ABC

from imspy.timstof.frame import TimsFrame
from imspy.timstof.slice import TimsSlice

import imspy_connector
ims = imspy_connector.py_dataset


class AcquisitionMode:
    def __init__(self, mode: str):
        """AcquisitionMode class.

        Args:
            mode (str): Acquisition mode.
        """
        allowed_modes = ["DDA", "DIA", "UNKNOWN", "PRECURSOR"]
        assert mode in allowed_modes, f"Unknown acquisition mode, use one of {allowed_modes}"
        self.__mode_ptr = ims.PyAcquisitionMode.from_string(mode)

    @property
    def mode(self) -> str:
        """Get the acquisition mode.

        Returns:
            str: Acquisition mode.
        """
        return self.__mode_ptr.acquisition_mode

    @classmethod
    def from_ptr(cls, ptr: ims.PyAcquisitionMode):
        """Get an AcquisitionMode from a pointer.

        Args:
            ptr (pims.AcquisitionMode): Pointer to an acquisition mode.

        Returns:
            AcquisitionMode: Acquisition mode.
        """
        instance = cls.__new__(cls)
        instance.__mode_ptr = ptr
        return instance

    def __repr__(self):
        return f"AcquisitionMode({self.mode})"


class TimsDataset(ABC):
    def __init__(self, data_path: str):
        """TimsDataHandle class.

        Args:
            data_path (str): Path to the data.
        """
        self.__dataset = None
        self.binary_path = None

        self.data_path = data_path
        self.meta_data = self.__load_meta_data()
        self.global_meta_data = self.__load_global_meta_data()
        self.precursor_frames = self.meta_data[self.meta_data["MsMsType"] == 0].Id.values.astype(np.int32)
        self.fragment_frames = self.meta_data[self.meta_data["MsMsType"] > 0].Id.values.astype(np.int32)
        self.__current_index = 1

        # Try to load the data with the first binary found
        appropriate_found = False
        for so_path in obb.get_so_paths():
            try:
                self.__dataset = ims.PyTimsDataset(self.data_path, so_path)
                self.binary_path = so_path
                appropriate_found = True
                break
            except Exception:
                continue
        assert appropriate_found is True, ("No appropriate bruker binary could be found, please check if your "
                                           "operating system is supported by open-tims-bruker-bridge.")

    @property
    def acquisition_mode(self) -> str:
        """Get the acquisition mode.

        Returns:
            str: Acquisition mode.
        """
        return self.__dataset.get_acquisition_mode()

    @property
    def acquisition_mode_numeric(self) -> int:
        """Get the acquisition mode as a numerical value.

        Returns:
            int: Acquisition mode as a numerical value.
        """
        return self.__dataset.get_acquisition_mode_numeric()

    @property
    def frame_count(self) -> int:
        """Get the number of frames.

        Returns:
            int: Number of frames.
        """
        return self.__dataset.frame_count()

    def __load_meta_data(self) -> pd.DataFrame:
        """Get the meta data.

        Returns:
            pd.DataFrame: Meta data.
        """
        return pd.read_sql_query("SELECT * from Frames", sqlite3.connect(self.data_path + "/analysis.tdf"))

    def __load_global_meta_data(self) -> Dict[str, str]:
        """Get the global meta data.

        Returns:
            pd.DataFrame: Global meta data.
        """
        d = pd.read_sql_query("SELECT * from GlobalMetadata", sqlite3.connect(self.data_path + "/analysis.tdf"))
        return dict(zip(d.Key, d.Value))

    @property
    def im_lower(self):
        return float(self.global_meta_data["OneOverK0AcqRangeLower"])

    @property
    def im_upper(self):
        return float(self.global_meta_data["OneOverK0AcqRangeUpper"])

    @property
    def mz_lower(self):
        return float(self.global_meta_data["MzAcqRangeLower"])

    @property
    def mz_upper(self):
        return float(self.global_meta_data["MzAcqRangeUpper"])

    @property
    def average_cycle_length(self) -> float:
        return np.mean(np.diff(self.meta_data.Time.values))

    @property
    def description(self) -> str:
        return self.global_meta_data["Description"]

    def get_tims_frame(self, frame_id: int) -> TimsFrame:
        """Get a TimsFrame.

        Args:
            frame_id (int): Frame ID.

        Returns:
            TimsFrame: TimsFrame.
        """
        return TimsFrame.from_py_tims_frame(self.__dataset.get_frame(frame_id))

    def get_tims_slice(self, frame_ids: NDArray[np.int32]) -> TimsSlice:
        """Get a TimsFrame.

        Args:
            frame_ids (int): Frame ID.

        Returns:
            TimsFrame: TimsFrame.
        """
        return TimsSlice.from_py_tims_slice(self.__dataset.get_slice(frame_ids))

    def tof_to_mz(self, frame_id: int, tof_values: NDArray[np.int32]) -> NDArray[np.float64]:
        """Convert TOF values to m/z values.

        Args:
            frame_id (int): Frame ID.
            tof_values (NDArray[np.int32]): TOF values.

        Returns:
            NDArray[np.float64]: m/z values.
        """
        return self.__dataset.tof_to_mz(frame_id, tof_values)

    def mz_to_tof(self, frame_id: int, mz_values: NDArray[np.float64]) -> NDArray[np.int32]:
        """Convert m/z values to TOF values.

        Args:
            frame_id (int): Frame ID.
            mz_values (NDArray[np.float64]): m/z values.

        Returns:
            NDArray[np.int32]: TOF values.
        """
        return self.__dataset.mz_to_tof(frame_id, mz_values)

    def scan_to_inverse_mobility(self, frame_id: int, scan_values: NDArray[np.int32]) -> NDArray[np.float64]:
        """Convert scan values to inverse mobility values.

        Args:
            frame_id (int): Frame ID.
            scan_values (NDArray[np.int32]): Scan values.

        Returns:
            NDArray[np.float64]: Inverse mobility values.
        """
        return self.__dataset.scan_to_inverse_mobility(frame_id, scan_values)

    def inverse_mobility_to_scan(self, frame_id: int, im_values: NDArray[np.float64]) -> NDArray[np.int32]:
        """Convert inverse mobility values to scan values.

        Args:
            frame_id (int): Frame ID.
            im_values (NDArray[np.float64]): Inverse mobility values.

        Returns:
            NDArray[np.int32]: Scan values.
        """
        return self.__dataset.inverse_mobility_to_scan(frame_id, im_values)

    def compress_zstd(self, values: NDArray[np.uint8]) -> NDArray[np.uint8]:
        """Compress values using ZSTD.

        Args:
            values (NDArray[np.float64]): Values to compress.

        Returns:
            NDArray[np.uint8]: Compressed values.
        """
        return self.__dataset.compress_bytes_zstd(values)

    def decompress_zstd(self, values: NDArray[np.uint8], ignore_first_n: int = 8) -> NDArray[np.uint8]:
        """Decompress values using ZSTD.

        Args:
            values (NDArray[np.float64]): Values to decompress.
            ignore_first_n (int): Number of bytes to ignore.

        Returns:
            NDArray[np.uint8]: Decompressed values.
        """
        return self.__dataset.decompress_bytes_zstd(values[ignore_first_n:])

    def indexed_values_to_compressed_bytes(self, scan_values: NDArray[np.int32], tof_values: NDArray[np.int32],
                                intensity_values: NDArray[np.float64], total_scans: int) -> NDArray[np.uint8]:
        """Convert scan and intensity values to bytes.

        Args:
            scan_values (NDArray[np.int32]): Scan values.
            tof_values (NDArray[np.int32]): TOF values.
            intensity_values (NDArray[np.float64]): Intensity values.
            total_scans (int): Total number of scans.

        Returns:
            NDArray[np.uint8]: Bytes.
        """
        return self.__dataset.scan_tof_intensities_to_compressed_u8(
            scan_values, 
            tof_values,
            intensity_values.astype(np.int32),
            total_scans
        )

    def compress_frames(self, frames: List[TimsFrame],
                                  total_scans: int, num_threads: int = 4) -> List[NDArray[np.uint8]]:
        """Compress a collection of frames.

        Args:
            frames (List[TimsFrame]): List of frames.
            total_scans (int): Total number of scans.
            num_threads (int): Number of threads to use.

        Returns:
            List[NDArray[np.uint8]]: List of compressed bytes.
        """
        return self.__dataset.compress_frames([f.get_frame_ptr() for f in frames], total_scans, num_threads)

    def bytes_to_indexed_values(self, values: NDArray[np.uint8]) \
            -> (NDArray[np.int32], NDArray[np.int32], NDArray[np.float64]):
        """Convert bytes to scan, tof, and intensity values.

        Args:
            values (NDArray[np.uint8]): Bytes.

        Returns:
            NDArray[np.int32]: Scan values.
            NDArray[np.int32]: TOF values.
            NDArray[np.float64]: Intensity values.
        """
        scan_values, tof_values, intensity_values = self.__dataset.u8_to_scan_tof_intensities(values)
        return scan_values, tof_values, intensity_values.astype(np.float64)

    def __iter__(self):
        return self

    def __next__(self):
        if self.__current_index <= self.frame_count:
            frame_ptr = self.__dataset.get_frame(self.__current_index)
            self.__current_index += 1
            if frame_ptr is not None:
                return TimsFrame.from_py_tims_frame(frame_ptr)
            else:
                raise ValueError(f"Frame pointer is None for valid index: {self.__current_index}")
        else:
            self.__current_index = 1  # Reset for next iteration
            raise StopIteration

    def __getitem__(self, index):
        if isinstance(index, slice):
            return self.get_tims_slice(np.arange(index.start, index.stop, index.step).astype(np.int32))
        return self.get_tims_frame(index)

    def __repr__(self):
        return f"TimsDataset({self.data_path})"



import plotly.graph_objects as go
from ipywidgets import widgets
import numpy as np
import abc
import plotly.express as px


def calculate_mz_tick_spacing(mz_min, mz_max, num_ticks=10):
    return np.round((mz_max - mz_min) / num_ticks, 1)


class ImsPointCloudVisualizer(abc.ABC):
    def __init__(self, data):
        self.data = data
        self.__create_widgets()

    def __create_widgets(self):
        points = np.array([[1.0, 1.0, 1.0, 1.0]])
        point_cloud = go.Scatter3d(
            x=points[:, 0],
            y=points[:, 1],
            z=points[:, 2],
            mode='markers',
            marker=dict(size=5, color=np.log(points[:, 3]), colorscale='inferno', opacity=1))

        self.points_widget = go.FigureWidget(data=[point_cloud])
        self.points_widget.update_layout(margin=dict(l=0, r=0, b=0, t=0),
                                         scene={'xaxis': {'title': 'X'},
                                                'yaxis': {'title': 'Y'},
                                                'zaxis': {'title': 'Z', 'dtick': 1}},
                                         width=800,  # Set the width here
                                         height=600,  # Set the height here
                                         template="plotly_white")

        self.opacity_slider = widgets.FloatSlider(value=0.5, min=0.1, max=1, step=0.1, description='opacity:',
                                                  continuous_update=False)

        self.point_size_slider = widgets.FloatSlider(value=1, min=0.1, max=5.0, step=0.1, description='point size:',
                                                     continuous_update=False)

        self.color_scale = widgets.Dropdown(options=sorted(px.colors.named_colorscales()),
                                            value='inferno', description='color scale:', disabled=False)

        self.point_controls = widgets.HBox(children=[self.opacity_slider, self.point_size_slider, self.color_scale])

        self.update_button = widgets.Button(description='Update')
        self.update_button.on_click(self.on_update_clicked)

        self.box = widgets.VBox(children=[self.point_controls, self.update_button, self.points_widget])

    @abc.abstractmethod
    def on_update_clicked(self, change):
        """
        """
        pass

    @abc.abstractmethod
    def display_widgets(self):
        """
        """
        pass


class DDAPrecursorPointCloudVis(ImsPointCloudVisualizer, abc.ABC):
    def __init__(self, data):
        super().__init__(data)
        self.opacity_slider.value = 0.3
        self.point_size_slider.value = 0.5

    def on_update_clicked(self, change):

        points = self.data
        f = np.sort(np.unique(np.copy(points.frame.values)))
        f_idx = dict(np.c_[f, np.arange(f.shape[0])])
        rt_indices = [f_idx[x] for x in points.frame.values]

        self.points_widget.data[0].x = rt_indices
        self.points_widget.data[0].y = points.scan.values
        self.points_widget.data[0].z = points.mz.values
        self.points_widget.data[0].marker = dict(size=self.point_size_slider.value,
                                                 color=np.log(points.intensity.values + 0.0001),
                                                 colorscale=self.color_scale.value,
                                                 line=dict(width=0),
                                                 opacity=self.opacity_slider.value)

        tick_spacing = calculate_mz_tick_spacing(np.min(points.mz.values), np.max(points.mz.values))

        self.points_widget.update_layout(margin=dict(l=0, r=0, b=0, t=0),
                                         scene={'xaxis': {'title': 'Rt-Index'},
                                                'yaxis': {'title': 'Mobility-Index'},
                                                'zaxis': {'title': 'm/z', 'dtick': tick_spacing}},
                                         template="plotly_white")

    def display_widgets(self):
        try:
            display(self.box)
        except Exception as e:
            print(e)



from abc import abstractmethod
import imspy_connector
ims = imspy_connector.py_quadrupole


class TimsTofCollisionEnergy:
    def __init__(self):
        pass

    @abstractmethod
    def get_collision_energy(self) -> float:
        pass

    @abstractmethod
    def get_collision_energies(self) -> list[float]:
        pass


class TimsTofCollisionEnergyDIA(TimsTofCollisionEnergy):
    def __init__(self, frame: list[int], frame_window_group: list[int], window_group: list[int], scan_start: list[int],
                 scan_end: list[int], collision_energy: list[float]):

        super().__init__()
        self.__ptr = ims.PyTimsTofCollisionEnergyDIA(frame, frame_window_group, window_group,
                                                     scan_start, scan_end, collision_energy)

    def get_collision_energy(self, frame_id: int, scan_id: int) -> float:
        return self.__ptr.get_collision_energy(frame_id, scan_id)

    def get_collision_energies(self, frame_ids: list[int], scan_ids: list[int]) -> list[float]:
        return self.__ptr.get_collision_energies(frame_ids, scan_ids)




import imspy_connector
ims = imspy_connector.py_chemistry


def one_over_k0_to_ccs(one_over_k0, mz, charge, mass_gas=28.013, temp=31.85, t_diff=273.15):
    """
    convert reduced ion mobility (1/k0) to CCS
    :param one_over_k0: reduced ion mobility
    :param charge: charge state of the ion
    :param mz: mass-over-charge of the ion
    :param mass_gas: mass of drift gas
    :param temp: temperature of the drift gas in C°
    :param t_diff: factor to translate from C° to K
    """
    return ims.one_over_reduced_mobility_to_ccs(one_over_k0, mz, charge, mass_gas, temp, t_diff)


def ccs_to_one_over_k0(ccs, mz, charge, mass_gas=28.013, temp=31.85, t_diff=273.15):
    """
    convert CCS to 1 over reduced ion mobility (1/k0)
    :param ccs: collision cross-section
    :param charge: charge state of the ion
    :param mz: mass-over-charge of the ion
    :param mass_gas: mass of drift gas (N2)
    :param temp: temperature of the drift gas in C°
    :param t_diff: factor to translate from C° to K
    """
    return ims.ccs_to_one_over_reduced_mobility(ccs, mz, charge, mass_gas, temp, t_diff)



import pandas as pd
import tensorflow as tf
from abc import ABC, abstractmethod
from numpy.typing import NDArray

from imspy.algorithm.utility import get_model_path
from imspy.utility import tokenize_unimod_sequence
from imspy.simulation.utility import irt_to_rts_numba


def load_deep_retention_time() -> tf.keras.models.Model:
    """ Get a pretrained deep predictor model

    Returns:
        The pretrained deep predictor model
    """
    return tf.keras.models.load_model(get_model_path('DeepRetentionTimePredictor'))


class PeptideChromatographyApex(ABC):
    """
    ABSTRACT INTERFACE for a chromatographic separation for peptides
    """

    def __init__(self):
        pass

    @abstractmethod
    def simulate_separation_times(self, sequences: list[str]) -> NDArray:
        pass

    @abstractmethod
    def simulate_separation_times_pandas(self, data: pd.DataFrame) -> pd.DataFrame:
        pass


class GRURetentionTimePredictor(tf.keras.models.Model):

    def __init__(self,
                 num_tokens,
                 seq_len=50,
                 emb_dim=128,
                 gru_1=128,
                 gru_2=64,
                 rdo=0.0,
                 do=0.2):
        super(GRURetentionTimePredictor, self).__init__()

        self.emb = tf.keras.layers.Embedding(input_dim=num_tokens + 1, output_dim=emb_dim, input_length=seq_len)

        self.gru1 = tf.keras.layers.Bidirectional(tf.keras.layers.GRU(gru_1, return_sequences=True,
                                                                      name='GRU1'))

        self.gru2 = tf.keras.layers.Bidirectional(tf.keras.layers.GRU(gru_2, return_sequences=False,
                                                                      name='GRU2',
                                                                      recurrent_dropout=rdo))

        self.dense1 = tf.keras.layers.Dense(128, activation='relu', name='Dense1',
                                            kernel_regularizer=tf.keras.regularizers.l1_l2(1e-3, 1e-3))

        self.dense2 = tf.keras.layers.Dense(64, activation='relu', name='Dense2',
                                            kernel_regularizer=tf.keras.regularizers.l1_l2(1e-3, 1e-3))

        self.dropout = tf.keras.layers.Dropout(do, name='Dropout')

        self.out = tf.keras.layers.Dense(1, activation=None, name='Output')

    def call(self, inputs):
        """
        :param inputs: should contain: (sequence)
        """
        # get inputs
        seq = inputs
        # sequence learning
        x_recurrent = self.gru2(self.gru1(self.emb(seq)))
        # regularize
        d1 = self.dropout(self.dense1(x_recurrent))
        # output
        return self.out(self.dense2(d1))


class DeepChromatographyApex(PeptideChromatographyApex):

    def __init__(self, model: GRURetentionTimePredictor, tokenizer: tf.keras.preprocessing.text.Tokenizer,
                 name: str = 'gru_predictor', verbose: bool = False):
        super(DeepChromatographyApex, self).__init__()
        self.model = model
        self.tokenizer = tokenizer
        self.name = name
        self.verbose = verbose

    def _preprocess_sequences(self, sequences: list[str], pad_len: int = 50) -> NDArray:
        char_tokens = [tokenize_unimod_sequence(seq) for seq in sequences]
        char_tokens = self.tokenizer.texts_to_sequences(char_tokens)
        char_tokens = tf.keras.preprocessing.sequence.pad_sequences(char_tokens, pad_len, padding='post')
        return char_tokens

    def simulate_separation_times(self, sequences: list[str], batch_size: int = 1024) -> NDArray:
        tokens = self._preprocess_sequences(sequences)
        tf_ds = tf.data.Dataset.from_tensor_slices(tokens).batch(batch_size)

        return self.model.predict(tf_ds, verbose=self.verbose)

    def simulate_separation_times_pandas(self, data: pd.DataFrame,
                                         gradient_length: float,
                                         batch_size: int = 1024) -> pd.DataFrame:
        tokens = self._preprocess_sequences(data.sequence.values)
        tf_ds = tf.data.Dataset.from_tensor_slices(tokens).batch(batch_size)

        irts = self.model.predict(tf_ds, verbose=self.verbose)
        rts = irt_to_rts_numba(irts, new_max=gradient_length)
        data[f'retention_time_{self.name}'] = rts
        return data



import numpy as np
import pandas as pd
import tensorflow as tf
from abc import ABC, abstractmethod
from numpy.typing import NDArray
from imspy.chemistry.mobility import ccs_to_one_over_k0
from scipy.optimize import curve_fit
from imspy.utility import tokenize_unimod_sequence
from imspy.algorithm.utility import get_model_path


def load_deep_ccs_predictor() -> tf.keras.models.Model:
    """ Get a pretrained deep predictor model

    Returns:
        The pretrained deep predictor model
    """
    return tf.keras.models.load_model(get_model_path('IonmobPredictor'))


class PeptideIonMobilityApex(ABC):
    """
    ABSTRACT INTERFACE for simulation of ion-mobility apex value
    """

    def __init__(self):
        pass

    @abstractmethod
    def simulate_ion_mobilities(self, sequences: list[str], charges: list[int]) -> NDArray:
        pass

    @abstractmethod
    def simulate_ion_mobilities_pandas(self, data: pd.DataFrame) -> pd.DataFrame:
        pass


def get_sqrt_slopes_and_intercepts(
        mz: np.ndarray,
        charge: np.ndarray,
        ccs: np.ndarray,
        fit_charge_state_one: bool = False
) -> (np.ndarray, np.ndarray):
    """

    Args:
        mz:
        charge:
        ccs:
        fit_charge_state_one:

    Returns:

    """

    if fit_charge_state_one:
        slopes, intercepts = [], []
    else:
        slopes, intercepts = [0.0], [0.0]

    if fit_charge_state_one:
        c_begin = 1
    else:
        c_begin = 2

    for c in range(c_begin, 5):
        def fit_func(x, a, b):
            return a * np.sqrt(x) + b

        triples = list(filter(lambda x: x[1] == c, zip(mz, charge, ccs)))

        mz_tmp, charge_tmp = np.array([x[0] for x in triples]), np.array([x[1] for x in triples])
        ccs_tmp = np.array([x[2] for x in triples])

        popt, _ = curve_fit(fit_func, mz_tmp, ccs_tmp)

        slopes.append(popt[0])
        intercepts.append(popt[1])

    return np.array(slopes, np.float32), np.array(intercepts, np.float32)


class ProjectToInitialSqrtCCS(tf.keras.layers.Layer):
    """
    Simple sqrt regression layer, calculates ccs value as linear mapping from mz, charge -> ccs
    """

    def __init__(self, slopes, intercepts):
        super(ProjectToInitialSqrtCCS, self).__init__()
        self.slopes = tf.constant([slopes])
        self.intercepts = tf.constant([intercepts])

    def call(self, inputs):
        mz, charge = inputs[0], inputs[1]
        # since charge is one-hot encoded, can use it to gate linear prediction by charge state
        return tf.expand_dims(tf.reduce_sum((self.slopes * tf.sqrt(mz) + self.intercepts) * tf.squeeze(charge), axis=1),
                              1)


class GRUCCSPredictor(tf.keras.models.Model):
    """
    Deep Learning model combining initial linear fit with sequence based features, both scalar and complex
    """

    def __init__(self, slopes, intercepts, num_tokens,
                 seq_len=50,
                 emb_dim=128,
                 gru_1=128,
                 gru_2=64,
                 rdo=0.0,
                 do=0.2):
        super(GRUCCSPredictor, self).__init__()
        self.__seq_len = seq_len

        self.initial = ProjectToInitialSqrtCCS(slopes, intercepts)

        self.emb = tf.keras.layers.Embedding(input_dim=num_tokens + 1, output_dim=emb_dim, input_length=seq_len)

        self.gru1 = tf.keras.layers.Bidirectional(tf.keras.layers.GRU(gru_1, return_sequences=True,
                                                                      name='GRU1'))

        self.gru2 = tf.keras.layers.Bidirectional(tf.keras.layers.GRU(gru_2, return_sequences=False,
                                                                      name='GRU2',
                                                                      recurrent_dropout=rdo))

        self.dense1 = tf.keras.layers.Dense(128, activation='relu',
                                            kernel_regularizer=tf.keras.regularizers.l1_l2(1e-3, 1e-3))
        self.dense2 = tf.keras.layers.Dense(64, activation='relu',
                                            kernel_regularizer=tf.keras.regularizers.l1_l2(1e-3, 1e-3))

        self.dropout = tf.keras.layers.Dropout(do)

        self.out = tf.keras.layers.Dense(1, activation=None)

    def call(self, inputs, **kwargs):
        """
        :param inputs: should contain: (mz, charge_one_hot, seq_as_token_indices)
        """
        # get inputs
        mz, charge, seq = inputs[0], inputs[1], inputs[2]
        # sequence learning
        x_recurrent = self.gru2(self.gru1(self.emb(seq)))
        # concat to feed to dense layers
        concat = tf.keras.layers.Concatenate()([charge, x_recurrent])
        # regularize
        d1 = self.dropout(self.dense1(concat))
        d2 = self.dense2(d1)
        # combine simple linear hypotheses with deep part
        return self.initial([mz, charge]) + self.out(d2), self.out(d2)


class DeepPeptideIonMobilityApex(PeptideIonMobilityApex):
    def __init__(self, model: GRUCCSPredictor, tokenizer: tf.keras.preprocessing.text.Tokenizer,
                 verbose: bool = False,
                 name: str = 'gru_predictor'):
        super(DeepPeptideIonMobilityApex, self).__init__()
        self.model = model
        self.tokenizer = tokenizer
        self.name = name
        self.verbose = verbose

    def _preprocess_sequences(self, sequences: list[str], pad_len: int = 50) -> NDArray:
        char_tokens = [tokenize_unimod_sequence(sequence) for sequence in sequences]
        char_tokens = self.tokenizer.texts_to_sequences(char_tokens)
        char_tokens = tf.keras.preprocessing.sequence.pad_sequences(char_tokens, pad_len, padding='post')
        return char_tokens

    def simulate_ion_mobilities(
            self,
            sequences: list[str],
            charges: list[int],
            mz: list[float],
            batch_size: int = 1024
    ) -> NDArray:
        tokenized_sequences = self._preprocess_sequences(sequences)

        # prepare masses, charges, sequences
        m = np.expand_dims(mz, 1)
        charges_one_hot = tf.one_hot(np.array(charges) - 1, 4)

        ds = tf.data.Dataset.from_tensor_slices(((m, charges_one_hot, tokenized_sequences), np.zeros_like(mz))).batch(batch_size)
        ccs, _ = self.model.predict(ds, verbose=self.verbose)

        return np.array([ccs_to_one_over_k0(c, m, z) for c, m, z in zip(ccs, mz, charges)])

    def simulate_ion_mobilities_pandas(self, data: pd.DataFrame, batch_size: int = 1024) -> pd.DataFrame:
        tokenized_sequences = self._preprocess_sequences(data.sequence.values)

        # prepare masses, charges, sequences
        m = np.expand_dims(data.mz.values, 1)
        charges_one_hot = tf.one_hot(np.array(data.charge.values) - 1, 4)

        ds = tf.data.Dataset.from_tensor_slices(((m, charges_one_hot, tokenized_sequences),
                                                 np.zeros_like(m))).batch(batch_size)

        ccs, _ = self.model.predict(ds, verbose=self.verbose)

        data[f'inv_mobility_{self.name}'] = np.array([ccs_to_one_over_k0(c, m, z)
                                                      for c, m, z in zip(ccs, m, data.charge.values)])

        data = data[['peptide_id', 'sequence', 'charge', 'relative_abundance', f'inv_mobility_{self.name}']]

        return data

    def __repr__(self):
        return f'DeepPeptideIonMobilityApex(name={self.name}, model={self.model})'



from __future__ import annotations
from typing import Optional, Tuple, List
import numpy as np
from numpy.typing import NDArray
from numpy.typing import ArrayLike
from abc import ABC, abstractmethod

from imspy.data.spectrum import MzSpectrum
from imspy.utility.utilities import gaussian, exp_gaussian, normal_pdf
import numba
import pyopenms

from imspy.simulation.noise import detection_noise
import imspy_connector
ims = imspy_connector.py_chemistry

MASS_PROTON = 1.007276466621
MASS_NEUTRON = 1.00866491595


def simulate_precursor_spectrum(sequence: str, charge: int) -> MzSpectrum:
    return MzSpectrum.from_py_mz_spectrum(ims.simulate_precursor_spectrum(sequence, charge))


def simulate_precursor_spectra(sequences: NDArray, charges: NDArray, num_threads: int) -> List[MzSpectrum]:
    spectra = ims.simulate_precursor_spectra(sequences, charges, num_threads)
    return [MzSpectrum.from_py_mz_spectrum(x) for x in spectra]


@numba.jit(nopython=True)
def factorial(n: int):
    if n <= 0:
        return 1
    return n * factorial(n - 1)


# linear dependency of mass
@numba.jit(nopython=True)
def lam(mass: float, slope: float = 0.000594, intercept: float = -0.03091):
    """
    :param intercept:
    :param slope:
    :param mass:
    :return:
    """
    return slope * mass + intercept


@numba.jit(nopython=True)
def weight(mass: float, peak_nums: NDArray, normalize: bool = True):
    """
    :param peak_nums:
    :param mass:
    :param normalize:
    :return:
    """
    factorials = np.zeros_like(peak_nums)
    norm = 1
    for i, k in enumerate(peak_nums):
        factorials[i] = factorial(k)
    weights = np.exp(-lam(mass)) * np.power(lam(mass), peak_nums) / factorials
    if normalize:
        norm = weights.sum()
    return weights / norm


def get_pyopenms_weights(sequence: str, peak_nums: ArrayLike, generator: pyopenms.CoarseIsotopePatternGenerator):
    """
    Gets hypothetical intensities of isotopic peaks.

    :param peak_nums:
    :param sequence: Peptide sequence in proForma format.
    :type sequence: str
    :param generator: pyopenms generator for isotopic pattern calculation
    :type generator: pyopenms.CoarseIsotopePatternGenerator
    :return: List of isotopic peak intensities
    :rtype: List
    """

    n = peak_nums.shape[0]
    generator.setMaxIsotope(n)
    aa_sequence = sequence.strip("_")
    peptide = pyopenms.AASequence().fromString(aa_sequence.replace("[", "(").replace("]", ")"))
    formula = peptide.getFormula()
    distribution = generator.run(formula)
    intensities = [i.getIntensity() for i in distribution.getContainer()]
    return intensities


@numba.jit(nopython=True)
def iso(x: NDArray, mass: float, charge: float, sigma: float, amp: float, K: int, step_size: float,
        add_detection_noise: bool = False, mass_neutron: float = MASS_NEUTRON):
    """
    :param mass_neutron:
    :param x:
    :param mass:
    :param charge:
    :param sigma:
    :param amp:
    :param K:4
    :param step_size:
    :param add_detection_noise:
    :return:
    """
    k = np.arange(K)
    means = ((mass + mass_neutron * k) / charge).reshape((1, -1))
    weights = weight(mass, k).reshape((1, -1))
    intensities = np.sum(weights * normal_pdf(x.reshape((-1, 1)), means, sigma), axis=1) * step_size
    if add_detection_noise:
        return detection_noise(intensities * amp)
    else:
        return intensities * amp


@numba.jit(nopython=True)
def numba_generate_pattern(
        lower_bound: float,
        upper_bound: float,
        mass: float,
        charge: float,
        amp: float,
        k: int,
        sigma: float = 0.008492569002123142,
        resolution: int = 3) -> Tuple[ArrayLike, ArrayLike]:
    """
    :param lower_bound:
    :param upper_bound:
    :param mass:
    :param charge:
    :param amp:
    :param k:
    :param sigma:
    :param resolution:
    :return:
    """
    step_size = min(sigma / 10, 1 / np.power(10, resolution))
    size = int((upper_bound - lower_bound) // step_size + 1)
    mzs = np.linspace(lower_bound, upper_bound, size)
    intensities = iso(mzs, mass, charge, sigma, amp, k, step_size)

    return mzs + MASS_PROTON, intensities


def numba_ion_sampler(mass: float, charge: int, sigma: ArrayLike, k: int, ion_count: int, intensity_per_ion: int):
    sigma = np.atleast_1d(sigma)
    if sigma.size == 1:
        sigma = np.repeat(sigma, k)
    if sigma.size != k:
        raise ValueError("Sigma must be either length 1 or k")

    us = np.zeros(ion_count)
    devs_std = np.zeros(ion_count)
    for ion in range(ion_count):
        u = np.random.uniform()
        dev_std = np.random.normal()
        us[ion] = u
        devs_std[ion] = dev_std

    weights = weight(mass, np.arange(k)).cumsum()

    def _get_component(u: float, weights_cum_sum: ArrayLike = weights) -> int:
        for idx, weight_sum in enumerate(weights_cum_sum):
            if u < weight_sum:
                return idx

    comps = np.zeros_like(us)
    devs = np.zeros_like(us)
    for idx, u in enumerate(us):
        comp = _get_component(u)
        comps[idx] = comp
        devs[idx] = sigma[comp] * devs_std[idx]

    mz = ((comps * MASS_NEUTRON) + mass) / charge + MASS_PROTON + devs
    i = np.ones_like(mz) * intensity_per_ion
    return mz, i


@numba.jit(nopython=True)
def create_initial_feature_distribution(num_rt: int, num_im: int,
                                        rt_lower: float = -9,
                                        rt_upper: float = 18,
                                        im_lower: float = -4,
                                        im_upper: float = 4,
                                        distr_im=gaussian,
                                        distr_rt=exp_gaussian) -> np.array:
    I = np.ones((num_rt, num_im)).astype(np.float32)

    for i, x in enumerate(np.linspace(im_lower, im_upper, num_im)):
        for j, y in enumerate(np.linspace(rt_lower, rt_upper, num_rt)):
            I[j, i] *= (distr_im(x) * distr_rt(y))

    return I


class IsotopePatternGenerator(ABC):
    @abstractmethod
    def generate_pattern(self, mass: float, charge: int) -> Tuple[ArrayLike, ArrayLike]:
        pass

    @abstractmethod
    def generate_spectrum(self, mass: int, charge: int, min_intensity: int) -> MzSpectrum:
        pass


class AveragineGenerator(IsotopePatternGenerator):
    def __init__(self):
        super().__init__()
        self.default_abundance = 1e4

    def generate_pattern(self, mass: float, charge: int, k: int = 7,
                         amp: Optional[float] = None, resolution: float = 3,
                         min_intensity: int = 5) -> Tuple[ArrayLike, ArrayLike]:
        pass

    def generate_spectrum(self, mass: int, charge: int, min_intensity: int = 5, k: int = 7,
                          amp: Optional[float] = None, resolution: int = 3, centroided: bool = True) -> MzSpectrum:
        if amp is None:
            amp = self.default_abundance

        lb = mass / charge - .2
        ub = mass / charge + k + .2

        mz, i = numba_generate_pattern(
            lower_bound=lb,
            upper_bound=ub,
            mass=mass,
            charge=charge,
            amp=amp,
            k=k,
            resolution=resolution
        )

        # TODO: implement centroid method
        if centroided:
            return MzSpectrum(mz, i) \
                .to_resolution(resolution).filter(lb, ub, min_intensity) \
                .to_centroided(baseline_noise_level=max(min_intensity, 1), sigma=1 / np.power(10, resolution - 1))

        return MzSpectrum(mz, i).to_resolution(resolution).filter(lb, ub, min_intensity)

    @staticmethod
    def generate_spectrum_by_sampling(
            mass: float,
            charge: int,
            k: int = 7,
            sigma: ArrayLike = 0.008492569002123142,
            ion_count: int = 1000000,
            resolution: int = 3,
            intensity_per_ion: int = 1,
            centroided: bool = True,
            min_intensity: int = 5) -> MzSpectrum:

        assert 100 <= mass / charge <= 2000, f"m/z should be between 100 and 2000, was: {mass / charge}"
        mz, i = numba_ion_sampler(mass, charge, sigma, k, ion_count, intensity_per_ion)

        if centroided:
            return (MzSpectrum(mz, i)
                    .to_resolution(resolution).filter(-1, -1, min_intensity)
                    .to_centroided(baseline_noise_level=max(min_intensity, 1), sigma=1 / np.power(10, resolution - 1))
                    )
        return MzSpectrum(mz, i) \
            .to_resolution(resolution).filter(-1, -1, min_intensity)


def generate_isotope_pattern_rust(mass: float, charge: int, min_intensity: float = 150, k: int = 7, resolution: int = 3,
                                  centroid: bool = True):
    return MzSpectrum.from_py_mz_spectrum(ims.generate_precursor_spectrum(
        mass=mass,
        charge=charge,
        min_intensity=min_intensity,
        k=k,
        resolution=resolution,
        centroid=centroid
    ))


def generate_isotope_patterns_rust(masses: NDArray, charges: NDArray, min_intensity: float = 0, k: int = 7,
                                   resolution: int = 3, centroid: bool = True, num_threads: int = 4):
    return [MzSpectrum.from_py_mz_spectrum(x) for x in ims.generate_precursor_spectra(
        masses=masses,
        charges=charges,
        min_intensity=min_intensity,
        k=k,
        resolution=resolution,
        centroid=centroid,
        num_threads=num_threads,
    )]



import sqlite3
from imspy.timstof.data import TimsDataset
import pandas as pd

import imspy_connector
ims = imspy_connector.py_dia


class TimsDatasetDIA(TimsDataset):
    def __init__(self, data_path: str):
        super().__init__(data_path=data_path)
        self.__dataset = ims.PyTimsDatasetDIA(self.data_path, self.binary_path)

    @property
    def dia_ms_ms_windows(self):
        """Get PASEF meta data for DIA.

        Returns:
            pd.DataFrame: PASEF meta data.
        """
        return pd.read_sql_query("SELECT * from DiaFrameMsMsWindows",
                                 sqlite3.connect(self.data_path + "/analysis.tdf"))

    @property
    def dia_ms_ms_info(self):
        """Get DIA MS/MS info.

        Returns:
            pd.DataFrame: DIA MS/MS info.
        """
        return pd.read_sql_query("SELECT * from DiaFrameMsMsInfo",
                                 sqlite3.connect(self.data_path + "/analysis.tdf"))



from typing import Union, List

from imspy.data.spectrum import MzSpectrum
from imspy.timstof.frame import TimsFrame
from numpy.typing import NDArray

import imspy_connector
ims = imspy_connector.py_quadrupole


class TimsTofQuadrupoleDIA:
    def __init__(self, frame: NDArray, frame_window_group: NDArray, window_group: NDArray, scan_start: NDArray,
                 scan_end: NDArray, isolation_mz: NDArray, isolation_width: NDArray, k: float | None = None):
        self.handle = ims.PyTimsTransmissionDIA(
            frame, frame_window_group, window_group, scan_start, scan_end, isolation_mz, isolation_width, k
        )

    def apply_transmission(self, frame_id: int, scan_id: int, mz: NDArray) -> NDArray:
        return self.handle.apply_transmission(frame_id, scan_id, mz)

    def transmit_spectrum(self, frame_id: int, scan_id: int, spectrum: MzSpectrum,
                          min_probability: float | None = None) -> MzSpectrum:
        return MzSpectrum.from_py_mz_spectrum(self.handle.transmit_spectrum(frame_id, scan_id,
                                                                            spectrum.get_spec_ptr(), min_probability))

    def transmit_frame(self, frame: TimsFrame, min_probability: float | None = None) -> TimsFrame:
        return TimsFrame.from_py_tims_frame(self.handle.transmit_tims_frame(frame.get_frame_ptr(), min_probability))

    def frame_to_window_group(self, frame_id: int) -> int:
        return self.handle.frame_to_window_group(frame_id)

    def is_transmitted(self, frame_id: int, scan_id: int, mz: float, min_proba: float | None = None) -> bool:
        return self.handle.is_transmitted(frame_id, scan_id, mz, min_proba)

    def any_transmitted(self, frame_id: int, scan_id: int, mz: NDArray, min_proba: float | None = None) -> bool:
        return self.handle.any_transmitted(frame_id, scan_id, mz, min_proba)

    def transmit_ion(self, frame_ids: NDArray, scan_ids: NDArray, spectrum: MzSpectrum, min_probability: Union[float, None]) -> List[List[MzSpectrum]]:
        transmission_profile = self.handle.transmit_ion(frame_ids, scan_ids, spectrum.get_spec_ptr(), min_probability)
        result = []
        for i in enumerate(frame_ids):
            scan_list = []
            for j in enumerate(scan_ids):
                scan_list.append(MzSpectrum.from_py_mz_spectrum(transmission_profile[i][j]))
            result.append(scan_list)

        return result

    def is_precursor(self, frame_id: int) -> bool:
        return self.handle.is_precursor(frame_id)

    def isotopes_transmitted(
            self,
            frame_id: int,
            scan_id: int,
            mz_mono: float,
            mz: NDArray,
            min_proba: float | None = None
    ) -> tuple[float, list[tuple[float, float]]]:
        """
        Get the transmission probability for a list of isotopes
        Args:
            frame_id:
            scan_id:
            mz_mono:
            mz:
            min_proba:

        Returns:

        """
        return self.handle.isotopes_transmitted(frame_id, scan_id, mz_mono, mz, min_proba)



import numpy as np
import pandas as pd
from typing import List

from numpy.typing import NDArray
from tensorflow import sparse as sp

from imspy.utility.utilities import re_index_indices
from imspy.timstof.frame import TimsFrame, TimsFrameVectorized
from imspy.timstof.frame import TimsSpectrum

import imspy_connector
ims = imspy_connector.py_tims_slice


class TimsSlice:
    def __init__(self,
                 frame_id: NDArray[np.int32],
                 scan: NDArray[np.int32],
                 tof: NDArray[np.int32],
                 retention_time: NDArray[np.float64],
                 mobility: NDArray[np.float64],
                 mz: NDArray[np.float64],
                 intensity: NDArray[np.float64]):
        """Create a TimsSlice.

        Args:
            frame_id (NDArray[np.int32]): Frame ID.
            scan (NDArray[np.int32]): Scan.
            tof (NDArray[np.int32]): TOF.
            retention_time (NDArray[np.float64]): Retention time.
            mobility (NDArray[np.float64]): Mobility.
            mz (NDArray[np.float64]): m/z.
            intensity (NDArray[np.float64]): Intensity.
        """

        assert len(frame_id) == len(scan) == len(tof) == len(retention_time) == len(mobility) == len(mz) == len(
            intensity), "All arrays must have the same length."

        self.__slice_ptr = ims.PyTimsSlice(
            frame_id, scan, tof, retention_time, mobility, mz, intensity
        )
        self.__current_index = 0

    @classmethod
    def from_py_tims_slice(cls, tims_slice: ims.PyTimsSlice):
        """Create a TimsSlice from a PyTimsSlice.

        Args:
            tims_slice (pims.PyTimsSlice): PyTimsSlice to create the TimsSlice from.

        Returns:
            TimsSlice: TimsSlice created from the PyTimsSlice.
        """
        instance = cls.__new__(cls)
        instance.__slice_ptr = tims_slice
        instance.__current_index = 0
        return instance

    @classmethod
    def from_frames(cls, frames: List[TimsFrame]):
        """Create a TimsSlice from a list of TimsFrames.

        Args:
            frames (List[TimsFrame]): List of TimsFrames.

        Returns:
            TimsSlice: TimsSlice created from the list of TimsFrames.
        """
        return cls.from_py_tims_slice(ims.PyTimsSlice.from_frames([frame.get_frame_ptr() for frame in frames]))

    @property
    def first_frame_id(self) -> int:
        """First frame ID.

        Returns:
            int: First frame ID.
        """
        return self.__slice_ptr.first_frame_id

    @property
    def last_frame_id(self) -> int:
        """Last frame ID.

        Returns:
            int: Last frame ID.
        """
        return self.__slice_ptr.last_frame_id

    def __repr__(self):
        return f"TimsSlice({self.first_frame_id}, {self.last_frame_id})"

    @property
    def precursors(self):
        return TimsSlice.from_py_tims_slice(self.__slice_ptr.get_precursors())

    @property
    def fragments(self):
        return TimsSlice.from_py_tims_slice(self.__slice_ptr.get_fragments_dda())

    def filter(self, mz_min: float = 0.0, mz_max: float = 2000.0, scan_min: int = 0, scan_max: int = 1000,
               mobility_min: float = 0.0,
               mobility_max: float = 2.0,
               intensity_min: float = 0.0, intensity_max: float = 1e9, num_threads: int = 4) -> 'TimsSlice':
        """Filter the slice by m/z, scan and intensity.

        Args:
            mz_min (float): Minimum m/z value.
            mz_max (float): Maximum m/z value.
            scan_min (int, optional): Minimum scan value. Defaults to 0.
            scan_max (int, optional): Maximum scan value. Defaults to 1000.
            mobility_min (float, optional): Minimum inverse mobility value. Defaults to 0.0.
            mobility_max (float, optional): Maximum inverse mobility value. Defaults to 2.0.
            intensity_min (float, optional): Minimum intensity value. Defaults to 0.0.
            intensity_max (float, optional): Maximum intensity value. Defaults to 1e9.
            num_threads (int, optional): Number of threads to use. Defaults to 4.

        Returns:
            TimsSlice: Filtered slice.
        """
        return TimsSlice.from_py_tims_slice(
            self.__slice_ptr.filter_ranged(mz_min, mz_max, scan_min, scan_max, mobility_min, mobility_max,
                                           intensity_min, intensity_max, num_threads))

    def filter_by_type(self,
                       mz_min_ms1: float = 0,
                       mz_max_ms1: float = 2000,
                       scan_min_ms1: int = 0,
                       scan_max_ms1: int = 1000,
                       inv_mob_min_ms1: float = 0,
                       inv_mob_max_ms1: float = 2,
                       intensity_min_ms1: float = 0,
                       intensity_max_ms1: float = 1e9,
                       mz_min_ms2: float = 0,
                       mz_max_ms2: float = 2000,
                       scan_min_ms2: int = 0,
                       scan_max_ms2: int = 1000,
                       inv_mob_min_ms2: float = 0,
                       inv_mob_max_ms2: float = 2,
                       intensity_min_ms2: float = 0,
                       intensity_max_ms2: float = 1e9,
                       num_threads: int = 4) -> 'TimsSlice':

        """Filter the slice by m/z, scan and intensity, for MS1 and MS2 with different ranges.

        Args:
            mz_min_ms1 (float, optional): Minimum m/z value for MS1. Defaults to 0.
            mz_max_ms1 (float, optional): Maximum m/z value for MS1. Defaults to 2000.
            scan_min_ms1 (int, optional): Minimum scan value for MS1. Defaults to 0.
            scan_max_ms1 (int, optional): Maximum scan value for MS1. Defaults to 1000.
            inv_mob_min_ms1 (float, optional): Minimum inverse mobility value for MS1. Defaults to 0.
            inv_mob_max_ms1 (float, optional): Maximum inverse mobility value for MS1. Defaults to 2.
            intensity_min_ms1 (float, optional): Minimum intensity value for MS1. Defaults to 0.
            intensity_max_ms1 (float, optional): Maximum intensity value for MS1. Defaults to 1e9.
            mz_min_ms2 (float, optional): Minimum m/z value for MS2. Defaults to 0.
            mz_max_ms2 (float, optional): Maximum m/z value for MS2. Defaults to 2000.
            scan_min_ms2 (int, optional): Minimum scan value for MS2. Defaults to 0.
            scan_max_ms2 (int, optional): Maximum scan value for MS2. Defaults to 1000.
            inv_mob_min_ms2 (float, optional): Minimum inverse mobility value for MS2. Defaults to 0.
            inv_mob_max_ms2 (float, optional): Maximum inverse mobility value for MS2. Defaults to 2.
            intensity_min_ms2 (float, optional): Minimum intensity value for MS2. Defaults to 0.
            intensity_max_ms2 (float, optional): Maximum intensity value for MS2. Defaults to 1e9.
            num_threads (int, optional): Number of threads to use. Defaults to 4.

        Returns:
            TimsSlice: Filtered slice.
        """
        return TimsSlice.from_py_tims_slice(
            self.__slice_ptr.filter_ranged_ms_type_specific(
                mz_min_ms1, mz_max_ms1, scan_min_ms1, scan_max_ms1, inv_mob_min_ms1,
                inv_mob_max_ms1, intensity_min_ms1, intensity_max_ms1, mz_min_ms2, mz_max_ms2,
                scan_min_ms2, scan_max_ms2, inv_mob_min_ms2, inv_mob_max_ms2, intensity_min_ms2,
                intensity_max_ms2, num_threads))

    @property
    def frames(self) -> List[TimsFrame]:
        """Get the frames.

        Returns:
            List[TimsFrame]: Frames.
        """
        return [TimsFrame.from_py_tims_frame(frame) for frame in self.__slice_ptr.get_frames()]

    def to_resolution(self, resolution: int, num_threads: int = 4) -> 'TimsSlice':
        """Convert the slice to a given resolution.

        Args:
            resolution (int): Resolution.
            num_threads (int, optional): Number of threads to use. Defaults to 4.

        Returns:
            TimsSlice: Slice with given resolution.
        """
        return TimsSlice.from_py_tims_slice(self.__slice_ptr.to_resolution(resolution, num_threads))

    def to_windows(self, window_length: float = 10, overlapping: bool = True, min_num_peaks: int = 5,
                   min_intensity: float = 1, num_threads: int = 4) -> List[TimsSpectrum]:
        """Convert the slice to a list of windows.

        Args:
            window_length (float, optional): Window length. Defaults to 10.
            overlapping (bool, optional): Whether the windows should overlap. Defaults to True.
            min_num_peaks (int, optional): Minimum number of peaks in a window. Defaults to 5.
            min_intensity (float, optional): Minimum intensity of a peak in a window. Defaults to 1.
            num_threads (int, optional): Number of threads to use. Defaults to 1.

        Returns:
            List[MzSpectrum]: List of windows.
        """
        return [TimsSpectrum.from_py_tims_spectrum(spec) for spec in self.__slice_ptr.to_windows(
            window_length, overlapping, min_num_peaks, min_intensity, num_threads)]

    def to_dense_windows(self, window_length: float = 10, resolution: int = 1, overlapping: bool = True,
                         min_num_peaks: int = 5, min_intensity: float = 0.0, num_theads: int = 4) -> (
            tuple)[list[NDArray], list[NDArray], list[NDArray]]:

        DW = self.__slice_ptr.to_dense_windows(window_length, overlapping, min_num_peaks, min_intensity, resolution,
                                               num_theads)

        scan_list, window_indices_list, values_list = [], [], []

        for values, scans, bins, row, col in DW:
            W = np.reshape(values, (row, col))
            scan_list.append(scans)
            window_indices_list.append(bins)
            values_list.append(W)

        return scan_list, window_indices_list, values_list

    @property
    def df(self) -> pd.DataFrame:
        """Get the data as a pandas DataFrame.

        Returns:
            pd.DataFrame: Data.
        """
        columns = ['frame', 'scan', 'tof', 'retention_time', 'mobility', 'mz', 'intensity']
        return pd.DataFrame({c: v for c, v in zip(columns, self.__slice_ptr.to_arrays())})

    def __iter__(self):
        return self

    def __next__(self):
        if self.__current_index < self.__slice_ptr.frame_count:
            frame_ptr = self.__slice_ptr.get_frame_at_index(self.__current_index)
            self.__current_index += 1
            if frame_ptr is not None:
                return TimsFrame.from_py_tims_frame(frame_ptr)
            else:
                raise ValueError("Frame pointer is None for valid index.")
        else:
            self.__current_index = 0  # Reset for next iteration
            raise StopIteration

    def vectorized(self, resolution: int = 2, num_threads: int = 4) -> 'TimsSliceVectorized':
        """Get a vectorized version of the slice.

        Args:
            resolution (int, optional): Resolution. Defaults to 2.
            num_threads (int, optional): Number of threads to use. Defaults to 4.

        Returns:
            TimsSliceVectorized: Vectorized version of the slice.
        """
        return TimsSliceVectorized.from_vectorized_py_tims_slice(self.__slice_ptr.vectorized(resolution, num_threads))

    def get_tims_planes(self, tof_max_value: int = 400_000, num_chunks: int = 7, num_threads: int = 4) -> List[
        'TimsPlane']:
        return [TimsPlane.from_py_tims_plane(plane) for plane in
                self.__slice_ptr.to_tims_planes(tof_max_value, num_chunks, num_threads)]


class TimsSliceVectorized:
    def __init__(self):
        self.__slice_ptr = None
        self.__current_index = 0

    @classmethod
    def from_vectorized_py_tims_slice(cls, tims_slice: ims.PyTimsSliceVectorized):
        """Create a TimsSlice from a PyTimsSlice.

        Args:
            tims_slice (pims.PyTimsSlice): PyTimsSlice to create the TimsSlice from.

        Returns:
            TimsSlice: TimsSlice created from the PyTimsSlice.
        """
        instance = cls.__new__(cls)
        instance.__slice_ptr = tims_slice
        instance.__current_index = 0
        return instance

    @property
    def first_frame_id(self) -> int:
        """First frame ID.

        Returns:
            int: First frame ID.
        """
        return self.__slice_ptr.first_frame_id

    @property
    def last_frame_id(self) -> int:
        """Last frame ID.

        Returns:
            int: Last frame ID.
        """
        return self.__slice_ptr.last_frame_id

    @property
    def precursors(self):
        return TimsSlice.from_py_tims_slice(self.__slice_ptr.get_precursors())

    @property
    def fragments(self):
        return TimsSlice.from_py_tims_slice(self.__slice_ptr.get_fragments_dda())

    @property
    def frames(self) -> List[TimsFrameVectorized]:
        """Get the frames.

        Returns:
            List[TimsFrame]: Frames.
        """
        return [TimsFrameVectorized.from_py_tims_frame_vectorized(frame) for frame in
                self.__slice_ptr.get_vectorized_frames()]

    @property
    def df(self) -> pd.DataFrame:
        """Get the data as a pandas DataFrame.

        Returns:
            pd.DataFrame: Data.
        """
        columns = ['frame', 'scan', 'tof', 'retention_time', 'mobility', 'index', 'intensity']
        return pd.DataFrame({c: v for c, v in zip(columns, self.__slice_ptr.to_arrays())})

    def __iter__(self):
        return self

    def __next__(self):
        if self.__current_index < self.__slice_ptr.frame_count:
            frame_ptr = self.__slice_ptr.get_frame_at_index(self.__current_index)
            self.__current_index += 1
            if frame_ptr is not None:
                return TimsFrameVectorized.from_py_tims_frame_vectorized(frame_ptr)
            else:
                raise ValueError("Frame pointer is None for valid index.")
        else:
            self.__current_index = 0
            raise StopIteration

    def __repr__(self):
        return f"TimsSliceVectorized({self.first_frame_id}, {self.last_frame_id})"

    def get_tensor_repr(self, dense=True, zero_index=True, re_index=True, frame_max=None, scan_max=None,
                        index_max=None):

        frames, scans, _, _, _, indices, intensities = self.__slice_ptr.to_arrays()

        if zero_index:
            scans = scans - np.min(scans)
            frames = frames - np.min(frames)
            indices = indices - np.min(indices)

        if re_index:
            frames = re_index_indices(frames)

        if scan_max is None:
            m_s = np.max(scans) + 1
        else:
            m_s = scan_max + 1

        if index_max is None:
            m_i = np.max(indices) + 1
        else:
            m_i = index_max + 1

        if frame_max is None:
            m_f = np.max(frames) + 1
        else:
            m_f = frame_max + 1

        sv = sp.reorder(
            sp.SparseTensor(indices=np.c_[frames, scans, indices], values=intensities, dense_shape=(m_f, m_s, m_i)))

        if dense:
            return sp.to_dense(sv)
        else:
            return sv


class TimsPlane:
    def __init__(self):
        self.__plane_ptr = None

    @classmethod
    def from_py_tims_plane(cls, plane: ims.PyTimsPlane):
        """Create a TimsPlane from a PyTimsPlane.

        Args:
            plane (pims.PyTimsPlane): PyTimsPlane to create the TimsPlane from.

        Returns:
            TimsPlane: TimsPlane created from the PyTimsPlane.
        """
        instance = cls.__new__(cls)
        instance.__plane_ptr = plane
        return instance

    @property
    def mz_mean(self):
        return self.__plane_ptr.mz_mean

    @property
    def mz_std(self):
        return self.__plane_ptr.mz_std

    @property
    def tof_mean(self):
        return self.__plane_ptr.tof_mean

    @property
    def tof_std(self):
        return self.__plane_ptr.tof_std

    @property
    def frame_ids(self):
        return self.__plane_ptr.frame_ids

    @property
    def scans(self):
        return self.__plane_ptr.scans

    @property
    def intensities(self):
        return self.__plane_ptr.intensity

    @property
    def retention_times(self):
        return self.__plane_ptr.retention_times

    @property
    def mobilities(self):
        return self.__plane_ptr.mobilities

    @property
    def num_points(self):
        return len(self.frame_ids)

    @property
    def df(self):
        return pd.DataFrame({
            'frame': self.frame_ids,
            'scan': self.scans,
            'retention_time': self.retention_times,
            'mobility': self.mobilities,
            'intensity': self.intensities
        })

    def __repr__(self):
        return (f"TimsPlane(mz_mean: "
                f"{np.round(self.mz_mean, 4)}, "
                f"mz_std: {np.round(self.mz_std, 4)},"
                f" tof_mean: {np.round(self.tof_mean, 4)}, "
                f"tof_std: {np.round(self.tof_std, 4)}, "
                f"num_points: {len(self.frame_ids)})")



import json
import math
import toml
import numpy as np
import importlib.resources as resources

from typing import List, Tuple, Any, Dict
from numpy.typing import NDArray

from importlib.abc import Traversable

from numba import jit
from scipy.stats import norm

from imspy.chemistry.constants import MASS_WATER
from imspy.chemistry.amino_acids import AMINO_ACID_MASSES
from imspy.chemistry.utility import calculate_mz

import imspy_connector
ims = imspy_connector.py_chemistry


@jit(nopython=True)
def add_uniform_noise(rt_abu: np.ndarray, noise_level: float = 1.0):
    noise = np.random.uniform(0, noise_level, len(rt_abu))
    noise_relative = (noise * rt_abu)
    noised_signal = rt_abu + noise_relative
    return (noised_signal / np.sum(noised_signal)) * np.sum(rt_abu)


@jit(nopython=True)
def flatten_prosit_array(array):
    array_return = np.zeros(174)

    ptr = 0

    for c in range(3):
        i_y = array[:, 0, c]
        array_return[ptr:ptr + 29] = i_y
        ptr += 29

        i_b = array[:, 1, c]
        array_return[ptr:ptr + 29] = i_b
        ptr += 29

    return array_return


def sequences_to_all_ions(
        sequences: List[str],
        charges: List[int],
        intensities_flat: List[List[float]],
        normalized: bool = True,
        half_charge_one: bool = True,
        num_threads: int = 4) -> List[str]:
    """
    Simulate ion intensities for a list of peptide sequences, charges, and collision energies.
    Args:
        sequences: List of peptide sequences
        charges: List of peptide charges
        intensities_flat: List of intensities
        normalized: Whether to normalize the intensities
        half_charge_one: Whether to divide the intensity by 2 if the charge is 1
        num_threads: Number of threads to use for the calculation

    Returns:
        NDArray: Array of ion intensities
    """
    return ims.sequence_to_all_ions_par(sequences, charges, intensities_flat, normalized, half_charge_one, num_threads)


def sequence_to_all_ions(
        sequence: str,
        charge: int,
        intensities_flat: List[float],
        normalized: bool = True,
        half_charge_one: bool = True) -> List[str]:
    """
    Simulate ion intensities for a peptide sequence, charge, and collision energy.
    Args:
        sequence: Peptide sequence
        charge: Peptide charge
        intensities_flat: List of intensities
        normalized: Whether to normalize the intensities
        half_charge_one: Whether to divide the intensity by 2 if the charge is 1

    Returns:
        NDArray: Array of ion intensities
    """
    return ims.sequence_to_all_ions_ims(sequence, charge, intensities_flat, normalized, half_charge_one)


def get_acquisition_builder_resource_path(acquisition_mode: str = 'dia') -> Traversable:
    """ Get the path to a pretrained model

    Args:
        acquisition_mode: The name of the model to load

    Returns:
        The path to the pretrained model
    """
    acquisition_mode = acquisition_mode.lower()
    assert acquisition_mode in ['dia', 'midia', 'slice', 'synchro'], \
        f"acquisition_mode needs to be one of 'dia', 'midia', 'slice', 'synchro', was: {acquisition_mode}"

    return resources.files('imspy.simulation.resources.configs').joinpath(acquisition_mode + 'pasef.toml')


def get_ms_ms_window_layout_resource_path(acquisition_mode: str) -> Traversable:
    """ Get the path to a pretrained model

    Returns:
        The path to the pretrained model
    """

    assert acquisition_mode in ['dia', 'midia', 'slice', 'synchro'], \
        f"acquisition_mode needs to be one of 'dia', 'midia', 'slice', 'synchro', was: {acquisition_mode}"

    return resources.files('imspy.simulation.resources.configs').joinpath(acquisition_mode + '_ms_ms_windows.csv')


def read_acquisition_config(acquisition_name: str = 'dia') -> Dict[str, Any]:

    file_path = get_acquisition_builder_resource_path(acquisition_name)

    with open(file_path, 'r') as config_file:
        config_data = toml.load(config_file)
    return config_data


# Function to convert a list (or a pandas series) to a JSON string
def python_list_to_json_string(lst, as_float=True, num_decimals: int = 4) -> str:
    if as_float:
        return json.dumps([float(np.round(x, num_decimals)) for x in lst])
    return json.dumps([int(x) for x in lst])


# load peptides and ions
def json_string_to_python_list(json_string):
    return json.loads(json_string)


def sequence_to_numpy(sequence: str, max_length: int = 30) -> NDArray:
    """
    translate a peptide sequence given as python string into a numpy array of characters with a fixed length
    Args:
        sequence: the peptide sequence
        max_length: the maximum length a sequence can have

    Returns:
        numpy array of characters
    """
    arr = np.full((1, max_length), fill_value="", dtype=object)
    for i, char in enumerate(sequence):
        arr[0, i] = char
    return np.squeeze(arr)


def calculate_b_y_fragment_mz(sequence: str, modifications: NDArray, is_y: bool = False, charge: int = 1) -> float:
    """
    Calculate the m/z value of a b or y fragment.
    Args:
        sequence: the peptide sequence
        modifications: potential modifications
        is_y: is the fragment a y ion
        charge: the charge state of the peptide precursor

    Returns:
        m/z value of the fragment
    """
    # return mz of empty sequence
    if len(sequence) == 0:
        return 0.0

    # add up raw amino acid masses and potential modifications
    mass = np.sum([AMINO_ACID_MASSES[s] for s in sequence]) + np.sum(modifications)

    # if sequence is n-terminal, add water mass and calculate mz
    if is_y:
        return calculate_mz(mass + MASS_WATER, charge)

    # otherwise, calculate mz
    return calculate_mz(mass, charge)


def calculate_b_y_ion_series_ims(sequence: str, modifications: NDArray, charge: int = 1) -> Tuple[List, List]:
    """
    Calculate the b and y ion series for a given peptide sequence.
    Args:
        sequence: the peptide sequence
        modifications: potential modifications
        charge: the charge state of the peptide precursor

    Returns:
        b ion series, y ion series
    """
    return ims.calculate_b_y_ion_series(sequence, modifications, charge)




def get_native_dataset_path(ds_name: str = 'NATIVE.d') -> str:
    """ Get the path to a pretrained model

    Args:
        ds_name: The name of the dataset to load

    Returns:
        The path to the pretrained model
    """
    return str(resources.files('imspy.simulation.resources').joinpath(ds_name))


@jit(nopython=True)
def generate_events(n, mean, min_val, max_val):
    generated_values = np.random.exponential(scale=mean, size=n * 10)
    filtered_values = np.empty(n, dtype=np.float32)
    count = 0

    for val in generated_values:
        if min_val <= val <= max_val:
            filtered_values[count] = float(int(val))
            count += 1
            if count == n:
                break

    return filtered_values[:count]


@jit(nopython=True)
def custom_cdf(x, mean, std_dev):
    """
    Custom implementation of the CDF for a normal distribution.
    """
    z = (x - mean) / std_dev
    return 0.5 * (1 + math.erf(z / math.sqrt(2)))


@jit(nopython=True)
def accumulated_intensity_cdf_numba(sample_start, sample_end, mean, std_dev):
    """
    Calculate the accumulated intensity between two points using the custom CDF.
    """
    cdf_start = custom_cdf(sample_start, mean, std_dev)
    cdf_end = custom_cdf(sample_end, mean, std_dev)
    return cdf_end - cdf_start


@jit(nopython=True)
def irt_to_rts_numba(irt: NDArray, new_min=0, new_max=120):
    """
    Scale an array of values from the original range (min_val, max_val) to a new range (new_min, new_max).

    Parameters
    ----------
    irt : NDArray
        Array of values to be scaled.
    new_min : float
        Minimum value of the new range.
    new_max : float
        Maximum value of the new range.

    Returns
    -------
    NDArray
        Array of scaled values.
    """
    min_val = np.min(irt)
    max_val = np.max(irt)
    scaled_values = new_min + (irt - min_val) * (new_max - new_min) / (max_val - min_val)
    return scaled_values


@jit(nopython=True)
def calculate_number_frames(gradient_length: float = 90 * 60, rt_cycle_length: float = .108) -> int:
    """ Calculate the number of frames that will be taken during the acquisition

    Parameters
    ----------
    gradient_length : float
        Length of the gradient in seconds
    rt_cycle_length : float
        Length of the RT cycle in seconds

    Returns
    -------
    int
        Number of frames that will be taken during the acquisition
    """
    return int(gradient_length / rt_cycle_length)


@jit(nopython=True)
def calculate_mobility_spacing(mobility_min: float, mobility_max: float, num_scans=500) -> float:
    """ Calculate the amount of mobility that will be occupied by a single scan

    Parameters
    ----------
    mobility_min : float
        Minimum mobility value
    mobility_max : float
        Maximum mobility value
    num_scans : int
        Number of scans that will be taken during the acquisition

    Returns
    -------
    float
        Mobility spacing
    """
    length = float(mobility_max) - float(mobility_min)
    return length / num_scans


def get_z_score_for_percentile(target_score=0.95):
    assert 0 <= target_score <= 1.0, f"target_score needs to be between 0 and 1, was: {target_score}"
    return norm.ppf(1 - (1 - target_score) / 2)


@jit(nopython=True)
def calculate_bounds_numba(mean, std, z_score):
    """
    Calculate the bounds of a normal distribution for a given z-score.

    Parameters
    ----------
    mean : float
        Mean of the normal distribution
    std : float
        Standard deviation of the normal distribution
    z_score : float
        Z-score of the normal distribution

    Returns
    -------
    float
        Lower bound of the normal distribution
    """
    return mean - z_score * std, mean + z_score * std


@jit(nopython=True)
def get_frames_numba(rt_value, times_array, std_rt, z_score):
    """
    Get the frames that will be acquired for a given retention time value.
    Parameters
    ----------
    rt_value : float
        Retention time value
    times_array : NDArray
        Array of retention times
    std_rt : float
        Standard deviation of the retention time
    z_score : float
        Z-score of the normal distribution

    Returns
    -------
    NDArray
        Array of frame indices
    """
    rt_min, rt_max = calculate_bounds_numba(rt_value, std_rt, z_score)
    first_frame = np.argmin(np.abs(times_array - rt_min)) + 1
    last_frame = np.argmin(np.abs(times_array - rt_max)) + 1
    rt_frames = np.arange(first_frame, last_frame + 1)

    return rt_frames


@jit(nopython=True)
def get_scans_numba(im_value, ims_array, scans_array, std_im, z_score):
    """
    Get the scans that will be acquired for a given ion mobility value.
    """
    im_min, im_max = calculate_bounds_numba(im_value, std_im, z_score)
    im_start = np.argmin(np.abs(ims_array - im_max))
    im_end = np.argmin(np.abs(ims_array - im_min))

    scan_start = scans_array[im_start]
    scan_end = scans_array[im_end]

    # Generate scan indices in the correct order given the inverse relationship
    im_scans = np.arange(scan_start, scan_end + 1)
    return im_scans



import pandas as pd

import imspy_connector
ims = imspy_connector.py_simulation


class TimsTofSyntheticsDataHandleRust:
    def __init__(self, path: str):
        self.path = path
        self.__handle = ims.PyTimsTofSyntheticsDataHandle(path)

    def get_py_ptr(self):
        return self.__handle

    def __repr__(self):
        return f"TimsTofSyntheticsDataHandleRust(path={self.path})"

    def get_transmitted_ions(self, num_threads: int = 4) -> pd.DataFrame:
        peptide_ids, ion_ids, sequences, charges, collision_energies = self.__handle.get_transmitted_ions(num_threads)
        return pd.DataFrame({
            'peptide_id': peptide_ids,
            'ion_id': ion_ids,
            'sequence': sequences,
            'charge': charges,
            'collision_energy': collision_energies
        })



import sqlite3
import os
from abc import ABC
from typing import List

import pandas as pd
from imspy.timstof.frame import TimsFrame

import imspy_connector
ims = imspy_connector.py_simulation


class TimsTofSyntheticFrameBuilderDIA:
    def __init__(self, db_path: str):
        self.handle = ims.PyTimsTofSyntheticsFrameBuilderDIA(db_path)

    def build_frame(self, frame_id: int, fragment: bool = True, isotope_fragments: bool = True) -> TimsFrame:
        frame = self.handle.build_frame(frame_id, fragment, isotope_fragments)
        return TimsFrame.from_py_tims_frame(frame)

    def build_frames(self, frame_ids: List[int], fragment: bool = True, num_threads: int = 4, isotope_fragments: bool = True) -> List[TimsFrame]:
        frames = self.handle.build_frames(frame_ids, fragment, num_threads, isotope_fragments)
        return [TimsFrame.from_py_tims_frame(frame) for frame in frames]

    def get_collision_energy(self, frame_id: int, scan_id: int) -> float:
        return self.handle.get_collision_energy(frame_id, scan_id)

    def get_collision_energies(self, frame_ids: List[int], scan_ids: List[int]) -> List[float]:
        return self.handle.get_collision_energies(frame_ids, scan_ids)


class TimsTofSyntheticPrecursorFrameBuilder:
    def __init__(self, db_path: str):
        self.handle = ims.PyTimsTofSyntheticsPrecursorFrameBuilder(db_path)

    def build_precursor_frame(self, frame_id: int):
        frame = self.handle.build_precursor_frame(frame_id)
        return TimsFrame.from_py_tims_frame(frame)

    def build_precursor_frames(self, frame_ids: List[int], num_threads: int = 4):
        frames = self.handle.build_precursor_frames(frame_ids, num_threads)
        return [TimsFrame.from_py_tims_frame(frame) for frame in frames]


class SyntheticExperimentDataHandle:
    def __init__(self,
                 database_path: str,
                 database_name: str = 'synthetic_data.db',
                 verbose: bool = True,
                 ):
        self.verbose = verbose
        self.base_path = database_path
        self.database_path = os.path.join(self.base_path, database_name)
        self.conn = None

        self._setup()

    def _setup(self):
        if not os.path.exists(self.base_path):
            if self.verbose:
                print(f"Creating data directory: {self.base_path}")
            os.makedirs(self.base_path)
        if self.verbose:
            print(f"Connecting to database: {self.database_path}")
        self.conn = sqlite3.connect(self.database_path)

    def create_table(self, table_name: str, table: pd.DataFrame):
        # Create a table from a pandas DataFrame
        table.to_sql(table_name, self.conn, if_exists='replace', index=False)

    def append_table(self, table_name: str, table: pd.DataFrame):
        # Append a table to an existing table in the database
        table.to_sql(table_name, self.conn, if_exists='append', index=False)

    def create_table_sql(self, sql):
        # Create a table as per the provided SQL statement
        try:
            cursor = self.conn.cursor()
            cursor.execute(sql)
            self.conn.commit()
        except sqlite3.Error as e:
            print(f"An error occurred: {e}")

    def close(self):
        # Close the database connection
        if self.conn:
            self.conn.close()

    def get_table(self, table_name: str) -> pd.DataFrame:
        # Get a table as a pandas DataFrame
        return pd.read_sql(f"SELECT * FROM {table_name}", self.conn)

    def list_tables(self):
        with self.conn:
            cursor = self.conn.cursor()
            cursor.execute("SELECT name FROM sqlite_master WHERE type='table';")
            tables = cursor.fetchall()
            return [table[0] for table in tables]

    def list_columns(self, table_name):
        if table_name not in self.list_tables():
            raise ValueError(f"Table '{table_name}' does not exist in the database.")

        with self.conn:
            cursor = self.conn.cursor()
            cursor.execute(f"PRAGMA table_info({table_name});")
            columns = cursor.fetchall()
            return [column[1] for column in columns]

    def __repr__(self):
        return f"SyntheticExperimentDataHandle(database_path={self.database_path})"


class SyntheticExperimentDataHandleDIA(SyntheticExperimentDataHandle, ABC):
    def __init__(self,
                 database_path: str,
                 database_name: str = 'synthetic_data.db',
                 verbose: bool = True,):
        super().__init__(database_path, database_name, verbose)
        self.dia_ms_ms_info = None
        self.dia_ms_ms_windows = None

        self._additional_setup()

    def _additional_setup(self):
        self.dia_ms_ms_info = self.get_table('dia_ms_ms_info')
        self.dia_ms_ms_windows = self.get_table('dia_ms_ms_windows')

    def get_frame_to_window_group(self):
        return dict(zip(self.dia_ms_ms_info.frame, self.dia_ms_ms_info.window_group))

    def get_window_group_settings(self):
        window_group_settings = {}

        for _, row in self.dia_ms_ms_windows.iterrows():
            key = (row.window_group, row.scan_start)
            value = (row.mz_mid, row.mz_width)
            window_group_settings[key] = value

        return window_group_settings


if __name__ == '__main__':

    # Example usage
    path = '/path/to/directory'
    db_name = 'experiment_data.db'
    handle = SyntheticExperimentDataHandle(path, db_name)

    # Create a table, for example
    sql_create_peptides_table = '''
    CREATE TABLE IF NOT EXISTS peptides (
        peptide_id INTEGER PRIMARY KEY,
        sequence TEXT NOT NULL,
        monoisotopic_mass REAL)
    '''
    handle.create_table(sql_create_peptides_table)

    # Close the connection when done
    handle.close()
